from agox.models.descriptors.ABC_descriptor import DescriptorBaseClass
from agox.models.descriptors.fingerprint import Fingerprint
from agox.models.descriptors.fingerprint_jax.fingerprint_jax import \
    JaxFingerprint
from agox.models.descriptors.simple_fingerprint import SimpleFingerprint
from agox.models.descriptors.soap import SOAP
from agox.models.descriptors.spectral_graph_descriptor import \
    SpectralGraphDescriptor
from agox.models.descriptors.voronoi import Voronoi
from agox.models.descriptors.voronoi_site import VoronoiSite



from typing import List, Optional, Tuple, Union

import numpy as np
from ase.atoms import Atoms
from ase.data import atomic_numbers
from ase.data.colors import jmol_colors
from matplotlib.colors import to_rgba, to_rgba_array
from numpy.typing import NDArray
from typing_extensions import Self

from agox.candidates import CandidateBaseClass
from agox.utils.plot.utils import plane_to_indices

# matplotlib 3.8.0 defines `maptplotlib.typing.ColorType`
ColorType = Union[
    Tuple[float, float, float],         # RGB
    Tuple[float, float, float, float],  # RGBA
    str                                 # hex, grayscale, color names, CN
]

IndicesType = Optional[Union[int, List[int], NDArray]]


class Colors:
    def __init__(self,
                 atoms: Union[Atoms, CandidateBaseClass],
                 color_palette: Optional[List[ColorType]] = None):
        """Container to store atomic colors for plotting purposes.

        The public methods of this class all return `self`, such that they can
        be chained. For example:
        `Colors(atoms).lighten(...).set_template('white')`

        Parameters
        ----------
        atoms : Union[Atoms, CandidateBaseClass]
            Atoms or Candidate object to store colors for.
        color_palette : Optional[List[ColorType]], optional
            Color palette to map elements to initial colors. If not set,
            `ase.data.colors.jmol_colors` is used.
        """

        self.atoms = atoms

        if color_palette is None:
            color_palette = jmol_colors

        self.colors: NDArray = to_rgba_array([color_palette[atom.number] for atom in self.atoms])

    def set_color(self,
                  color: ColorType,
                  *,
                  indices: IndicesType = None) -> Self:
        """Set the color of one or more atoms.

        Parameters
        ----------
        color : ColorType
            Color to set. This can be any matplotlib-supported color.
        indices : IndicesType, optional
            Atomic indices to set color for, by default None (all atoms).

        Returns
        -------
        Self
        """

        indices = self._get_indices(indices)

        self.colors[indices] = to_rgba(color)

        return self

    def set_alpha(self,
                  alpha: float,
                  *,
                  indices: IndicesType = None) -> Self:
        """Set the alpha value of one or more atoms.

        Parameters
        ----------
        alpha : float
            Alpha value to set.
        indices : IndicesType, optional
            Atomic indices to set color for, by default None (all atoms).

        Returns
        -------
        Self
        """

        indices = self._get_indices(indices)

        self.colors[indices, 3] = alpha

        return self

    def set_element(self,
                    element: Union[str, int],
                    color: ColorType) -> Self:
        """Set the color of all atoms of one element.

        Parameters
        ----------
        element : Union[str, int]
            Element to set color for. This can be a symbol or an atomic number.
        color : ColorType
            Color to set. This can be any matplotlib-supported color.

        Returns
        -------
        Self
        """

        if isinstance(element, str):
            element = atomic_numbers[element]

        self.set_color(color,
                       indices=np.flatnonzero(self.atoms.numbers == element))

        return self

    def set_template(self,
                     color: ColorType) -> Self:
        """Set the color of all atoms of the template. This requires that this
        object was instantiated with a Candidate object.

        Parameters
        ----------
        color : ColorType
            Color to set. This can be any matplotlib-supported color.

        Returns
        -------
        Self

        Raises
        ------
        ValueError
            Raised when this object was not instantiated with a Candidate
            object.
        """

        if not isinstance(self.atoms, CandidateBaseClass):
            raise ValueError('Atoms provided is not a Candidate object. '
                             'Use `set_color` and a list of indices for Atoms objects.')

        self.set_color(color,
                       indices=self.atoms.template_indices)

        return self

    def darken(self,
               *,
               factor: float = 0.1,
               threshold: Optional[float] = None,
               plane: str = 'xy+',
               indices: IndicesType = None) -> Self:
        """Darken atom colors by a given factor.

        If `threshold` is set, atom colors are darkened based on the distance
        from `threshold` to the atomic coordinate (in the direction pointing
        out of the figure), multiplied by `factor`.
        Otherwise, atom colors are darkened by a constant factor `factor`.

        Parameters
        ----------
        factor : float, optional
            Factor to darken atom colors by, by default 0.1.
        threshold : Optional[float], optional
            Threshold to darken atom colors from, by default None (constant
            darken factor).
        plane : str, optional
            Plane to project the atoms on. See `plane_to_indices` for
            information on valud plane strings, by default 'xy+'.
        indices : IndicesType, optional
            Atomic indices to darken color for, by default None (all atoms).

        Returns
        -------
        Self
        """

        indices = self._get_indices(indices)
        _, _, d2, kv = plane_to_indices(plane)

        factor_ = factor if threshold is None else \
            ((kv * self.atoms[indices].positions[:, d2] - threshold) * factor)[:, np.newaxis]
        self.colors[indices, :3] = self.colors[indices, :3] * np.minimum(1, 1 - np.minimum(1, factor_))

        return self

    def lighten(self,
                *,
                factor: float = 0.1,
                threshold: Optional[float] = None,
                plane: str = 'xy+',
                indices: IndicesType = None) -> Self:
        """Lighten atom colors by a given factor.

        If `threshold` is set, atom colors are lightened based on the distance
        from `threshold` to the atomic coordinate (in the direction pointing
        out of the figure), multiplied by `factor`.
        Otherwise, atom colors are lightened by a constant factor `factor`.

        Parameters
        ----------
        factor : float, optional
            Factor to lighten atom colors by, by default 0.1.
        threshold : Optional[float], optional
            Threshold to lighten atom colors from, by default None (constant
            lighten factor).
        plane : str, optional
            Plane to project the atoms on. See `plane_to_indices` for
            information on valud plane strings, by default 'xy+'.
        indices : IndicesType, optional
            Atomic indices to lighten color for, by default None (all atoms).

        Returns
        -------
        Self
        """

        indices = self._get_indices(indices)
        _, _, d2, kv = plane_to_indices(plane)

        factor_ = factor if threshold is None else \
            ((kv * self.atoms[indices].positions[:, d2] - threshold) * factor)[:, np.newaxis]
        self.colors[indices, :3] = 1 - (1 - self.colors[indices, :3]) * np.minimum(1, 1 - np.minimum(1, factor_))

        return self

    def _get_indices(self, indices: IndicesType) -> NDArray:
        """Internal method to convert `indices` arguments to a numpy array of
        indices.

        Parameters
        ----------
        indices : IndicesType
            Indices to convert; either `int`, list of `int`, or numpy array.

        Returns
        -------
        NDArray
            Numpy array of indices.
        """
        if isinstance(indices, int):
            return np.array([indices], dtype=int)
        elif indices is None:
            return np.arange(len(self.atoms))
        else:
            return np.array(indices, dtype=int)

    def __getitem__(self, index) -> NDArray:
        """Return the color of an atom or atoms.

        Parameters
        ----------
        index
            Index or indices of atom or atoms.

        Returns
        -------
        NDArray
            Color of atom or atoms.
        """
        return self.colors[index]



from ase.io import write

from agox.writer import agox_writer
from .ABC_sampler import SamplerBaseClass
import numpy as np
from sklearn.cluster import KMeans
from scipy.spatial.distance import cdist

from agox.utils.filters.kmeans_energy import KMeansEnergyFilter

class KMeansSampler(SamplerBaseClass):

    """
    KMeans Sampler for selecting candidates based on a KMeans clustering algorithm.

    Parameters
    ----------
    descriptor : agox.models.descriptor.ABC_descriptor.DescriptorBaseClass
        Descriptor to use for the similarity criterion.
    model : agox.models.model.ABC_model.ModelBaseClass
        Model to use for the similarity criterion.
    sample_size : int
        The number of sample members, or population size. 
    max_energy : float
        The maximum energy difference, compared to the lowest energy structure, 
        for a structure to be considered for sampling.
    """

    name = "SamplerKMeans"
    parameters = {}

    def __init__(
        self, descriptor=None, model=None, sample_size=10, max_energy=5, **kwargs
    ):
        super().__init__(**kwargs)
        self.descriptor = descriptor
        self.sample_size = sample_size
        self.max_energy = max_energy
        self.sample = []
        self.sample_features = []
        self.model = model
        self.debug = False

        energy_filter = KMeansEnergyFilter(sample_size=sample_size, max_energy=max_energy)
        if self.filters is None:
            self.filters = energy_filter
        else:
            from agox.utils.filters import SumFilter
            self.filters = SumFilter(f0=self.filters, f1=energy_filter)

    def setup(self, all_finished_structures):
        """
        Setup up the KMeans Sampler.

        Parameters
        ----------
        all_finished_structures : list of ase.Atoms or agox.candidates.candidate.Candidate
        """

        # Check if there are any finished structures
        if len(all_finished_structures) < 1:
            self.sample = []
            return False

        # Sort the structures according to energy:
        energies = np.array([s.get_potential_energy() for s in all_finished_structures])

        # Calculate features:
        X = self.get_features(all_finished_structures)

        # Determine the number of clusters:
        # TODO: Why is this necessary?
        n_clusters = 1 + min(self.sample_size - 1, int(np.floor(len(energies) / 5)))

        # Perform KMeans clustering:
        kmeans = KMeans(n_clusters=n_clusters, init="k-means++", n_init=10, 
                        random_state=np.random.randint(0, 10e6)).fit(X)
        labels = kmeans.labels_

        # For each cluster find the lowest energy member:
        sample_indices = self.select_from_clusters(energies, n_clusters, labels)
        sample = [all_finished_structures[i] for i in sample_indices]
        sample_features = [X[i] for i in sample_indices]
        cluster_centers = [kmeans.cluster_centers_[i] for i in labels[sample_indices]]

        # Sort the sample mmebers according to energy.
        # This is only important for the output.
        sample_energies = [t.get_potential_energy() for t in sample]
        sorting_indices = np.argsort(sample_energies)
        self.sample = [sample[i] for i in sorting_indices]
        self.sample_features = [sample_features[i] for i in sorting_indices]
        self.cluster_centers = [cluster_centers[i] for i in sorting_indices]

        # Print information about the sample:
        self.print_output(all_finished_structures)

        return True

    def select_from_clusters(self, energies, n_clusters, labels):
        """
        Select the lowest energy member from each cluster.

        Parameters
        ----------
        energies : list of float
            List of energies of the structures.
        n_clusters : int
            Number of clusters.
        labels : list of int
            List of cluster labels for each structure.

        Returns
        -------
        sample_indices : list of int
            List of indices of the lowest energy member of each cluster.
        """
        indices = np.arange(len(energies))
        sample_indices = []
        for n in range(n_clusters):
            filt_cluster = labels == n
            cluster_indices = indices[filt_cluster]
            if len(cluster_indices) == 0:
                continue
            min_e_index = np.argmin(energies[filt_cluster])
            index_best_in_cluster = cluster_indices[min_e_index]
            sample_indices.append(index_best_in_cluster)
        return sample_indices

    def get_label(self, candidate):
        """
        Get the label of a candidate.

        Parameters
        ----------
        candidate : ase.Atoms or agox.candidates.candidate.Candidate
            Object to get the label of.

        Returns
        -------
        label : int
            The label of the candidate.
        """

        if len(self.sample) == 0:
            return None

        # find out what cluster we belong to
        f_this = np.array(self.descriptor.get_global_features([candidate]))
        distances = cdist(f_this, self.cluster_centers, metric="euclidean").reshape(-1)

        label = int(np.argmin(distances))
        return label

    def get_closest_sample_member(self, candidate):
        """
        Get the sample member that the candidate belongs to.

        Parameters
        ----------
        """
        label = self.get_label(candidate)
        cluster_center = self.sample[label]
        return cluster_center

    def get_features(self, structures):
        features = np.array(self.descriptor.get_features(structures)).sum(axis=1)
        return features

    def print_output(self, all_finished_structures):
        sample_energies = [t.get_potential_energy() for t in self.sample]
        for i, sample_energy in enumerate(sample_energies):
            self.writer(f"{i}: Sample energy {sample_energy:8.3f}")

        if self.model is not None and self.model.ready_state:
            for s in self.sample:
                t = s.copy()
                t.set_calculator(self.model)
                E = t.get_potential_energy()
                sigma = self.model.get_property("uncertainty")
                s.add_meta_information("model_energy", E)
                s.add_meta_information("uncertainty", sigma)
            self.writer(
                "SAMPLE_MODEL_ENERGY",
                "[",
                ",".join(
                    [
                        "{:8.3f}".format(t.get_meta_information("model_energy"))
                        for t in self.sample
                    ]
                ),
                "]",
            )
            self.writer(
                "SAMPLE_MODEL_SIGMA",
                "[",
                ",".join(
                    [
                        "{:8.3f}".format(t.get_meta_information("uncertainty"))
                        for t in self.sample
                    ]
                ),
                "]",
            )

        if self.debug:
            write(
                f"filtered_strucs_iteration_{self.get_iteration_counter()}.traj",
                all_finished_structures,
            )
            write(f"sample_iteration_{self.get_iteration_counter()}.traj", self.sample)

    @property
    def initialized(self):
        return len(self) > 0


from abc import ABC, abstractmethod
from typing import List, Tuple

import numpy as np
from ase import Atoms


class FilterBaseClass(ABC):
    """Base class for filters.

    This class is used to define the base class for filters. All filters
    should inherit from this class and implement the methods defined here.

    Parameters
    ----------

    Methods
    -------
    _filter(atoms: List[Atoms]) -> np.ndarray:
        Filter the atoms.

    __call__(atoms: List[Atoms]) -> Tuple[List[Atoms], np.ndarray]:
        Filter the atoms and return the selected atoms and indexes of the filtered atoms.

    """

    def __init__(self, **kwargs):
        """Initialize the filter."""
        super().__init__(**kwargs)

    def filter(self, atoms: List[Atoms]) -> Tuple[List[Atoms], np.ndarray]:
        indexes = self._filter(atoms)
        return [atoms[i] for i in indexes], indexes

    @abstractmethod
    def _filter(self, atoms: List[Atoms]) -> np.ndarray:
        """Filter the atoms object.

        Parameters
        ----------
        atoms
            The atoms object to be filtered.

        Returns
        -------
        indexes: array-like
            The indexes of the atoms that are kept.
        """
        pass

    @property
    @abstractmethod
    def name(self) -> str:
        return NotImplementedError

    def __call__(self, atoms: List[Atoms]) -> List[Atoms]:
        return self.filter(atoms)

    def __add__(self, other: "FilterBaseClass"):
        return SumFilter(f0=self, f1=other)


class SumFilter(FilterBaseClass):
    """Sum filter.

    This class implement to sum of two filters.

    Parameters
    ----------
    f0 : FilterBaseClass
        The first filter.
    f1 : FilterBaseClass
        The second filter.

    """

    def __init__(self, f0: FilterBaseClass, f1: FilterBaseClass, **kwargs):
        """Initialize the filter."""
        super().__init__(**kwargs)
        self.f0 = f0
        self.f1 = f1

    def _filter(self, atoms: List[Atoms]) -> List[Atoms]:
        """Filter the atoms object.

        Parameters
        ----------
        atoms
            The atoms object to be filtered.

        Returns
        -------
        List[Atoms]
            The filtered atoms object.
        """
        f0_atoms, idx0 = self.f0(atoms)
        _, idx1 = self.f1(f0_atoms)
        return idx0[idx1]

    @property
    def name(self) -> str:
        return f"{self.f0.name} + {self.f1.name}"



import numpy as np

class Confinement:

    def __init__(self, confinement_cell=None, confinement_corner=None, indices=None, pbc=[False]*3):
        self.confinement_cell = np.array(confinement_cell)
        self.confinement_corner = np.array(confinement_corner)

        if indices is None:
            indices = np.array([])
        self.indices = np.array(indices).flatten()

        self.periodicity_setup(pbc)
        self.confined = self.confinement_cell is not None and self.confinement_corner is not None

    def periodicity_setup(self, pbc):
        if isinstance(pbc, bool):
            self.pbc = [pbc]*3
        elif len(pbc) == 3:
            self.pbc = list(pbc)
        else:
            print('pbc should be list or bool! Setting pbc=False.')
            self.pbc = [False]*3

        self.hard_boundaries = [not p for p in self.pbc]

        if np.any(self.pbc):
            periodic_cell_vectors = self.confinement_cell[:, self.pbc]
            non_periodic_cell_vectors = self.confinement_cell[:, self.hard_boundaries]
            if np.any(np.abs(np.matmul(periodic_cell_vectors.T, non_periodic_cell_vectors)) > 0):
                print('---- BOX CONSTRAINT ----')
                print('Periodicity does not work for non-square non-periodic directions!')
                print('------------------------')

        if np.all(self.confinement_cell[:, 2] == 0):
            self.dimensionality = 2
            self.effective_confinement_cell = self.confinement_cell[0:2, 0:2]
            if np.all(self.confinement_cell[:, 1] == 0):
                self.effective_confinement_cell = self.confinement_cell[0:1, 0:1]
                self.dimensionality = 1
        else:
            self.effective_confinement_cell = self.confinement_cell
            self.dimensionality = 3

    def get_projection_coefficients(self, positions):
        positions = positions.reshape(-1, 3)
        return np.linalg.solve(self.effective_confinement_cell.T, (positions-self.confinement_corner)[:, 0:self.dimensionality].T).T.reshape(-1, self.dimensionality)

    def check_confinement(self, positions):
        """
        Finds the fractional coordinates of the atomic positions in terms of the box defined by the constraint. 
        """
        if self.confined:        
            C = self.get_projection_coefficients(positions)
            inside = np.all((C > 0) * (C < 1), axis=1)
            return inside
        else:
            return np.ones(positions.shape[0]).astype(bool)

    def get_confinement_limits(self):
        """
        This returns the confinement-limit lists which always assumes a square box. 
        """
        conf = [self.confinement_corner, self.confinement_cell @ np.array([1, 1, 1]) + self.confinement_corner]
        return conf

    def _get_box_vector(self, cell, corner):
        return cell.T @ np.random.rand(3) + corner

    def set_confinement_cell(self, cell, confinement_corner):        
        self.confinement_cell = cell
        self.confinement_corner = confinement_corner
        self.confined = True
        self.confined = self.confinement_cell is not None and self.confinement_corner is not None

    def set_dimensionality(self, dimensionality):
        self.dimensionality = dimensionality

    



from agox.test.run_tests.run_utils import agox_test_run
import pytest

@pytest.mark.parametrize("mode", ["rss_cluster_emt", "rss_bulk_emt", "rss_surface_emt", "rss_2d"])
def test_rss(mode, tmp_path, cmd_options):
    agox_test_run(mode, tmp_path, cmd_options)




from copy import deepcopy

import numpy as np
from ase import Atoms
from ase.constraints import FixInternals
from ase.data import covalent_radii
from ase.geometry import get_distances
from scipy.spatial.distance import cdist

from agox.generators.ABC_generator import GeneratorBaseClass


class BlockGeneratorBaseClass(GeneratorBaseClass):
    def __init__(
        self, building_blocks, N_blocks, apply_constraint=False, use_mic=True, **kwargs
    ):
        super().__init__(**kwargs)

        self.building_blocks = building_blocks
        self.N_blocks = np.array(N_blocks)

        for building_block in self.building_blocks:
            building_block.positions -= building_block.get_center_of_mass()

        self.apply_constraint = apply_constraint
        if apply_constraint:
            self.fix_internal_constraints = []
            for i, block in enumerate(building_blocks):
                for constraint in block.constraints:
                    if isinstance(constraint, FixInternals):
                        self.fix_internal_constraints.append(constraint)
                    else:
                        raise NotImplementedError(
                            "Transferring constraints other than FixInternals from building blocks is not implemented."
                        )

                if len(self.fix_internal_constraints) != i + 1:
                    self.fix_internal_constraints.append(None)

        self.use_mic = use_mic

    def extract_block(self, candidate, block_indices):
        positions = candidate.positions[block_indices]
        numbers = candidate.numbers[block_indices]
        cell = candidate.get_cell()

        atoms = Atoms(numbers=numbers, positions=positions, cell=cell)

        return atoms

    def block_move(self, candidate, block, internal_indices):
        candidate.positions[internal_indices] = block.positions

    def compare_environment_and_blocks(self, environment):
        numbers_list = environment.get_numbers()

        if len(numbers_list) != len(
            np.repeat(self.building_block.get_atomic_numbers(), self.N)
        ):
            self.writer("Building Blocks and N does not match environment settings!")
            self.writer("Will break now!")
            exit()

    def check_distances(self, candidate, block, internal_indices=None, present=True):
        if len(candidate) == 0:
            return True

        if self.use_mic:
            return self.check_distances_mic(candidate, block, internal_indices, present)

        # Find the distances of relevant atoms - we mask out the block
        # itself if it is already present in the candidate object.
        mask = np.ones(len(candidate)).astype(bool)
        if present:
            mask[internal_indices] = False
        D = cdist(block.positions, candidate.positions[mask, :])

        # We want to figure out if these are within allowed distances.
        # We make a matrix of the sum of covalent_radii such that
        # cv_matrix[0, 0] is the sum of covalent radii between a
        # block atom and a candidate atom.
        N_block = len(block)
        N_candidate = np.count_nonzero(mask)
        block_numbers = np.repeat(
            block.numbers[:, np.newaxis], axis=1, repeats=N_candidate
        )
        candidate_numbers = np.repeat(
            candidate.numbers[mask][np.newaxis, :], axis=0, repeats=N_block
        )
        block_cv = covalent_radii[block_numbers]
        candidate_cv = covalent_radii[candidate_numbers]
        cv = block_cv + candidate_cv

        # First we check thhat all distances are over c1 * cv
        min_check = (self.c1 * cv < D).all()
        # Then check that ANY distance is smaller than c2 * cv
        max_check = (self.c2 * cv > D).any()
        # Combined this esnures that we at least have ONE bond of appropriate length
        # and that NONE are too short.
        return min_check * max_check

    def check_distances_mic(
        self, candidate, block, internal_indices=None, present=True
    ):
        # Find the distances of relevant atoms - we mask out the block
        # itself if it is already present in the candidate object.
        mask = np.ones(len(candidate)).astype(bool)
        if present:
            mask[internal_indices] = False
        _, D = get_distances(
            block.positions,
            candidate.positions[mask, :],
            cell=candidate.get_cell(),
            pbc=candidate.get_pbc(),
        )

        cv = np.add.outer(
            covalent_radii[block.numbers], covalent_radii[candidate.numbers[mask]]
        )
        min_check = (self.c1 * cv < D).all()
        max_check = (self.c2 * cv > D).any()

        return min_check * max_check

    def get_fix_internal_constraint(self, bbs_used):
        blocks_in_order = [self.building_blocks[index] for index in bbs_used]
        index_offsets = np.cumsum([0] + [len(bb) for bb in blocks_in_order])

        all_bonds = []
        all_angles = []
        all_dihedrals = []

        for bb_index, atom_index_offset in zip(bbs_used, index_offsets):
            if self.fix_internal_constraints[bb_index] is None:
                continue
            else:
                base_constraint = deepcopy(self.fix_internal_constraints[bb_index])

            # Get the base constraint list:
            bonds = base_constraint.bonds.copy()
            angles = base_constraint.angles.copy()
            dihedrals = base_constraint.dihedrals.copy()

            # Update the indices:
            for value_index_list in [bonds, angles, dihedrals]:
                for value_index in value_index_list:
                    value_index[1] += atom_index_offset

            all_bonds += bonds
            all_angles += angles
            all_dihedrals += dihedrals

        return FixInternals(bonds=all_bonds, angles=all_angles, dihedrals=all_dihedrals)

    def get_building_block(self, placed, random_rotation=True):
        # Get a building block:
        remaining = np.argwhere((self.N_blocks - placed) > 0).flatten()
        index = int(np.random.choice(remaining, size=1)[0])
        bb = self.building_blocks[index].copy()

        if random_rotation:
            phi0, phi1, phi2 = np.random.uniform(0, 360, size=3)
            bb.rotate(phi2, (0, 0, 1))
            bb.rotate(phi0, (1, 0, 0))
            bb.rotate(phi1, (0, 1, 0))

        return bb, index

    def start_candidate(self):
        """
        This method generates a candidate using the start generator, which allows other generators
        to kick-start the sampler.
        """
        from agox.generators.block_generators.random import \
            RandomBlockGenerator

        return RandomBlockGenerator(
            self.building_blocks,
            self.N_blocks,
            confinement_cell=self.confinement_cell,
            confinement_corner=self.confinement_corner,
            c1=self.c1,
            c2=self.c2,
            use_mic=self.use_mic,
        )(self.sampler, self.environment)



import numpy as np
from abc import ABC, abstractmethod
from agox.module import Module
from agox.candidates.ABC_candidate import CandidateBaseClass
from agox.utils import candidate_list_comprehension

class DescriptorBaseClass(ABC, Module):

    def __init__(self, environment, surname='', use_cache=True):
        Module.__init__(self, surname=surname, use_cache=use_cache)        
        self.environment = environment

        assert self.descriptor_type in ['global', 'local'], 'Descriptor type must be either "global" or "local".'

    ##########################################################################################################
    # Create methods - Implemented by classes that inherit from this base-class.
    # Not called directly by methods of other classes. 
    ##########################################################################################################

    @property
    @abstractmethod
    def descriptor_type(self):
        """
        Returns
        -------
        str
            The type of descriptor either 'global' or 'local'. 
            Generally this refers to the n_centers dimension of calculated features, 
            which is 1 for global descriptors and typically (but not required) n_atoms for local features. 
        """
        pass

    @property
    @abstractmethod
    def get_number_of_centers(self, atoms):
        """
        Returns
        -------
        int
            The number of centers that the descriptor will be calculated for. 
            This is the n_centers dimension of the calculated features. 
        """
        pass

    def create_features(self, atoms):
        """
        Method to implement on child classes that does the calculation 
        of feature vectors. 

        This method should not (generally) deal with being given a list of 
        Atoms objects. 

        Parameters
        ----------
        atoms : Atoms object

        Returns
        --------
        Features of shape [n_centers, n_features]
        """
        pass

    def create_feature_gradient(self, atoms):
        """
        Method to implement on child classes that does the calculation
        of feature gradients.

        This method should not (generally) deal with being given a list of 
        Atoms objects. 

        Parameters
        ----------
        atoms : Atoms object

        Returns
        --------
        Feature gradients of shape [n_centers, n_atoms, 3, n_features]
        """
        return self.central_difference_gradient(atoms)

    ##########################################################################################################
    # Get methods - Ones to use in other scripts.
    ##########################################################################################################

    @candidate_list_comprehension
    @CandidateBaseClass.cache('features')
    def get_features(self, atoms):
        """
        Method to get global features.

        Parameters
        ----------
        atoms : Atoms object or list or Atoms objects.
            Atoms to calculate features for. 

        Returns
        -------
        list
            Global features for the given atoms.
        """
        return self.create_features(atoms)


    @candidate_list_comprehension
    @CandidateBaseClass.cache('feature_gradient')
    def get_feature_gradient(self, atoms):
        """
        Method to get global features.

        Parameters
        ----------
        atoms : Atoms object or list or Atoms objects.
            Atoms to calculate features for. 

        Returns
        -------
        list
            Global feature gradients for the given atoms.
        """
        return self.create_feature_gradient(atoms)

    @Module.reset_cache_key
    def change_descriptor_somehow(self):
        """
        This is not a real method. 

        This is just to illustrate that if you use the caching capability of the 
        descriptor-baseclass then you MUST use the @Module.reset_cache_key decorator 
        on any function that changes the descriptor - e.g. changing parameters. 
        """
        return
    
    def central_difference_gradient(self, atoms, delta=0.0001):
        """
        Calculate the gradient of the features using a central difference method.
        """

        f0 = self.get_features(atoms)
        dF_dx = np.zeros((f0.shape[0], len(atoms), 3, f0.shape[1]))
        
        for i in range(len(atoms)):
            for d in range(3):
                atoms.positions[i, d] -= delta
                fm = self.get_features(atoms)
                atoms.positions[i, d] += 2 * delta
                fp = self.get_features(atoms)
                atoms.positions[i, d] -= delta
                dF_dx[:, i, d, :] = (fp - fm) / (2 * delta)

        return dF_dx

    def __eq__(self, other):
        if not isinstance(other, DescriptorBaseClass):
            return False
        return self.cache_key == other.cache_key



import numpy as np
from abc import ABC, abstractmethod

from agox.utils.constraints.box_constraint import BoxConstraint
from ase.constraints import FixAtoms

from agox.module import Module
from agox.candidates import StandardCandidate

class EnvironmentBaseClass(ABC, Module):

    """
    Base class for all environments.

    Environments are used to define the environment in which the candidates are
    generated. This includes confinement, constraints, and other settings.

    Parameters
    ----------
    confinement_cell : np.ndarray, optional
        The cell of the confinement, by default None.
    confinement_corner : np.ndarray, optional
        The corner of the confinement, by default None.
    constraints : list, optional
        List of constraints, by default [].
    use_box_constraint : bool, optional
        If True, a box constraint is used, by default True.
    box_constraint_pbc : list, optional
        List of booleans, by default [False]*3.
    fix_template : bool, optional
        If True, the template is fixed, by default True.
    """

    name = 'Environment'



    def __init__(self, confinement_cell=None, confinement_corner=None, constraints=[], 
                 use_box_constraint=True, box_constraint_pbc=[False]*3, fix_template=True, surname=''):
        Module.__init__(self, surname=surname)
                
        self.confinement_cell = confinement_cell
        self.confinement_corner = confinement_corner
        self.constraints = constraints 
        self.use_box_constraint = use_box_constraint
        self.box_constraint_pbc = box_constraint_pbc
        self.fix_template = fix_template
        
    @abstractmethod
    def get_template(self, **kwargs):
        pass

    @abstractmethod    
    def get_numbers(self, numbers, **kwargs):
        pass
        
    @abstractmethod
    def environment_report(self):
        pass 

    def get_missing_indices(self):
        return np.arange(len(self._template), len(self._template)+len(self._numbers))

    def get_confinement_cell(self, distance_to_edge=3):
        if self.confinement_cell is not None:
            confinement_cell = self.confinement_cell
        elif self._template.pbc.all() is False:
            confinement_cell = self._template.get_cell().copy() - np.eye(3) * distance_to_edge * 2
        else: 
            # Find the directions that are not periodic: 
            confinement_cell = self._template.get_cell().copy() - np.eye(3) * distance_to_edge * 2
            directions = np.argwhere(self._template.pbc == True)
            for d in directions:
                confinement_cell[d, :] = self._template.get_cell()[d, :]

        return confinement_cell

    def get_confinement_corner(self, distance_to_edge=3):
        if self.confinement_cell is not None:
            confinement_corner = self.confinement_corner
        elif self._template.pbc.all() is False:
            confinement_corner = np.ones(3) * distance_to_edge
        else: 
            # Find the directions that are not periodic: 
            confinement_corner = np.ones(3) * distance_to_edge
            directions = np.argwhere(self._template.pbc == True)
            for d in directions:
                confinement_corner[d] = 0
        return confinement_corner

    def get_confinement(self):
        return {'confinement_cell':self.confinement_cell, 'confinement_corner':self.confinement_corner}

    def get_box_constraint(self):
        confinement_cell = self.get_confinement_cell()
        confinement_corner = self.get_confinement_corner()
        return BoxConstraint(confinement_cell=confinement_cell, confinement_corner=confinement_corner,
                             indices=self.get_missing_indices(), pbc=self.box_constraint_pbc)

    def get_constraints(self):
        constraints = []
        if self.use_box_constraint:
            constraints += [self.get_box_constraint()]
        if self.fix_template:
            constraints += [self.get_template_constraint()]
        return constraints + self.constraints 

    def get_template_constraint(self):
        return FixAtoms(indices=np.arange(len(self.get_template())))

    def convert_to_candidate_object(self, template):
        template = StandardCandidate(template=template, positions=template.positions, numbers=template.numbers, 
                                     cell=template.cell, info=template.info)
        return template

    def plot(self, name='environment_plot.png'):
        import matplotlib.pyplot as plt
        from agox.utils.matplotlib_utils import use_agox_mpl_backend
        from agox.utils.plot import plot_atoms, plot_cell

        use_agox_mpl_backend()

        atoms = self.get_template()
        atoms.set_constraint(self.get_constraints())

        fig, axs = plt.subplots(1, 3, figsize=(15, 5))

        for ax, plane in zip(axs, ['xy', 'xz', 'yz']):
            plot_atoms(ax, atoms, plane=plane, plot_constraint=True, repeat=True)
            plot_cell(ax, atoms.cell, plane=plane,
                      collection_kwargs=dict(edgecolors='black', linestyles='dotted'))
            plot_cell(ax, self.confinement_cell, plane=plane, offset=self.confinement_corner,
                      collection_kwargs=dict(edgecolors='red', linestyles='dashed'))

            # Equal aspect ratio            
            ax.set_aspect('equal', 'box')

        fig.savefig(name, bbox_inches='tight')
        plt.close()


import numpy as np
from agox.models.descriptors.ABC_descriptor import DescriptorBaseClass

def eval_gauss(r, x, width):
    return np.exp(-0.5*((x-r)/width)**2)

class SimpleFingerprint(DescriptorBaseClass):

    name = 'SimpleFingerprint'
    descriptor_type = 'local'

    def __init__(self, Nbins=30, width=0.2, r_cut=3, separate_center_species=True, *args, **kwargs):
        super().__init__(*args, **kwargs)
        
        self.Nbins = Nbins
        self.width = width
        self.r_cut = r_cut
        self.separate_center_species = separate_center_species

        species = self.environment.get_species()
        self.Nspecies = len(species)
        self.species_dict = {species[i]: i for i in range(len(species))}
        
        self.r_bins = np.linspace(0,self.r_cut, Nbins)

    def create_features(self, atoms):
        Natoms = len(atoms)
        symbols = atoms.get_chemical_symbols()

        distances = atoms.get_all_distances(mic=True)
        
        if self.separate_center_species:
            descriptor = np.zeros((Natoms, self.Nspecies, self.Nspecies, self.Nbins))
        else:
            descriptor = np.zeros((Natoms, self.Nspecies, self.Nbins))
        for i, (distances_i, sym_i) in enumerate(zip(distances, symbols)):
            for j, (r_ij, sym_j) in enumerate(zip(distances_i, symbols)):
                if i == j:
                    continue
                if r_ij > self.r_cut + 4*self.width:
                    continue 
                idx_sym_i = self.species_dict[sym_i]
                idx_sym_j = self.species_dict[sym_j]
                if self.separate_center_species:
                    descriptor[i, idx_sym_i, idx_sym_j] += 0.5*eval_gauss(r_ij, self.r_bins, self.width) / r_ij**2
                else:
                    descriptor[i, idx_sym_j] += 0.5*eval_gauss(r_ij, self.r_bins, self.width) / r_ij**2
        return descriptor.reshape(Natoms, -1)

    def create_global_features(self, atoms):
        return self.create_local_features(atoms).sum(axis=0)

    def get_number_of_centers(self, atoms):
        return len(atoms)

    @classmethod
    def from_species(cls, species, **kwargs):
        from ase import Atoms
        from agox.environments import Environment
        environment = Environment(template=Atoms(''), symbols=''.join(species), use_box_constraint=False,
                                  print_report=False)
        return cls(environment=environment, **kwargs)
        
    



from abc import ABC, abstractmethod
import numpy as np
from ase.io import write
from agox.candidates.ABC_candidate import CandidateBaseClass
from agox.module import Module
from agox.observer import Observer
from agox.writer import Writer, agox_writer

class SamplerBaseClass(ABC, Observer, Writer):

    """
    Base class for samplers.

    Parameters
    ----------
    database : database object
        An instance of an AGOX database class which the sampler will attach to, 
        and retrieve candidates from.
    filters : agox.utils.filters.FilterBaseClass
        Filter object that will be used to filter candidates from the database.
    sets : dict
        Dictionary with the sets that the sampler will observe.
    gets : dict
        Dictionary with the gets that the sampler will observe.
    order : int
        Order of the observer.
    verbose : bool    
    """
    
    def __init__(self, database=None, filters=None, sets={}, gets={}, order=1, verbose=True, use_counter=True, prefix='', 
        use_transfer_data=True, surname=''):
        Observer.__init__(self, sets=sets, gets=gets, order=order, surname=surname)
        Writer.__init__(self, verbose=verbose, use_counter=use_counter, prefix=prefix)
        self.sample = []
        self.use_transfer_data = use_transfer_data
        self.transfer_data = []
        self.filters = filters

        self.add_observer_method(self.setup_sampler,
                                 sets=self.sets[0], gets=self.gets[0], order=self.order[0],
                                 handler_identifier='database')

        if database is not None:
            self.attach_to_database(database)

    ########################################################################################
    # Required properties
    ########################################################################################    

    @property
    @abstractmethod
    def name(self):
        return NotImplementedError

    @property
    def initialized(self):
        """
        Property that returns True if the sampler has been initialized, such 
        that it is able to return a sample member. 

        Can be overwritten if the sampler has a different way of checking if
        it is initialized.
        """
        return len(self) > 0
    
    ########################################################################################
    # Required methods
    ########################################################################################

    @abstractmethod
    def setup(self, all_candidates):
        """

        Function that does the setup for the Sampler, e.g. filters candidates 
        from the database or calculates probabilities. 

        Must set self.sample.

        Parameters
        ----------
        all_candidates : list
            List of all candidates that will be considered by the sampler. If 
            'setup_sampler' is not overwritten this will be all candidates 
            currently in the database during in an AGOX run.
        """
        return None

    ########################################################################################
    # Default methods
    ########################################################################################
    
    @agox_writer
    @Observer.observer_method
    def setup_sampler(self, database, state):
        """
        Observer function that is attached to the database. The database passes 
        itself and the state. 

        The state is not passed to the 'self.setup' function that is called, so 
        if any data on there is required it can be set as an attribute, e.g. 
            self.something = state.get(...)
        or (probably preferably) the this function can be overwritten and more 
        things can be parsed to sampler.setup(). 

        Parameters
        ----------
        database : database object. 
            An instance of an AGOX database class. 
        state : state object
            An instance of the AGOX State object. 

        Returns
        -------
        state object
            The state object with any changes the sampler has made. 
        """
        database_data = database.get_all_candidates()
        if self.filters is not None:
            database_data, indices = self.filters(database_data)

        all_candidates = database_data + self.transfer_data
        
        if self.do_check():
            self.setup(all_candidates)
        
    def get_parents(self, number_of_parents = 1, replace = True):
        """
        Method that returns a list of parents for the generator.

        Parameters
        ----------
        number_of_parents : int
            Number of parents to return
        replace : bool
            If the parents should be sampled with replacement or not.

        Returns
        -------
        list
            List of parents.
        """
        if len(self.sample) == 0:
            return []
        elif len(self.sample) == number_of_parents:
            return self.get_all_members()

        indices = np.random.choice(len(self.sample), size=number_of_parents, replace=replace)
        parents = []
        for index in indices:
            member = self.sample[index].copy()
            self.sample[index].copy_calculator_to(member)
            parents.append(member)
        return parents

    def get_all_members(self):
        members = []
        for i, member in enumerate(self.sample):
            member = member.copy()
            self.sample[i].copy_calculator_to(member)
            members.append(member)
        return members

    def add_transfer_data(self, data):
        """
        Add structures to be considered by the Sampler, such that these structures 
        are passed to 'self.setup' when it is called.
        
        The added structures need a few things:
        1. To be Candidate objects.
        2. To have a energies.

        Feature-update: Should check that the data matches a given environment.

        Parameters
        ----------
        data : list
            List of candidate or atoms objects to be considered by the sampler
        """
        correct_type = np.array([isinstance(dat, CandidateBaseClass) for dat in data]).all()
        if not correct_type:
            raise TypeError('''Only candidate objects can be specified as transfer data, you probably gave ASE Atoms objects.''')
        self.transfer_data = data

    def set_sample(self, data):
        """
        Sets the sample to the given list of candidates, can be used to
        initialize the sample to something specific - e.g. for basin-hopping
        with the MetropolisSampler.
        When using other samplers (such as KMeansSampler) these may be
        overwritten and not considered when making the next sample - if that is
        not the wanted behaviour then use 'add_transfer_data' instead/aswell.

        The added structures need a few things:
        1. To be Candidate objects.
        2. To have a energies.    

        Parameters
        ----------
        data : list
            List of candidate objects.
        """
        correct_type = np.array([isinstance(dat, CandidateBaseClass) for dat in data]).all()
        if not correct_type:
            raise TypeError('''Only candidate objects can be specified as transfer data, you probably gave ASE Atoms objects.''')
        self.sample = data

    def __len__(self):
        return len(self.sample)

    def attach_to_database(self, database):
        from agox.databases.ABC_database import DatabaseBaseClass
        assert isinstance(database, DatabaseBaseClass)
        print(f'{self.name}: Attaching to database: {database}')
        self.attach(database)



from agox.models.descriptors.ABC_descriptor import DescriptorBaseClass
from agox.models.descriptors.fingerprint_cython.angular_fingerprintFeature_cy import \
    Angular_Fingerprint
import numpy as np

class Fingerprint(DescriptorBaseClass):

    name = "Fingerprint"
    descriptor_type = 'global'

    """
    Fingerprint descriptor that calculates radial and angular distribution functions 
    for a given atoms object. 

    Parameters
    ----------
    rc1 : float
        Radial cutoff for the radial distribution function.
    rc2 : float
        Radial cutoff for the angular distribution function.
    binwidth : float
        Binwidth for the radial distribution function.
    Nbins : int
        Number of bins for the angular distribution function.
    sigma1 : float
        Width of the gaussian used for the radial distribution function.
    sigma2 : float
        Width of the gaussian used for the angular distribution function.
    gamma : float
        Parameter of the cutoff function. 
    eta : float
        Weighting factor for the angular distribution function. 
    """

    def __init__(
        self,
        rc1=6,
        rc2=4,
        binwidth=0.2,
        Nbins=30,
        sigma1=0.2,
        sigma2=0.2,
        gamma=2,
        eta=20,
        use_angular=True,
        *args,
        **kwargs
    ):
        super().__init__(*args, **kwargs)

        self.cython_module = Angular_Fingerprint(
            self.environment.get_atoms(),
            Rc1=rc1,
            Rc2=rc2,
            binwidth1=binwidth,
            Nbins2=Nbins,
            sigma1=sigma1,
            sigma2=sigma2,
            gamma=gamma,
            eta=eta,
            use_angular=use_angular,
        )

    def create_features(self, atoms) -> np.ndarray:
        """

        Calculate feature vector for an atoms object.

        Parameters
        ----------
        atoms : Atoms object
        
        Returns
        -------
        features : np.ndarray
        """
        return self.cython_module.get_feature(atoms).reshape(1, -1)

    def create_feature_gradient(self, atoms) -> np.ndarray:
        """
        Calculate derivative of features with respect to atomic positions for
        given atoms object.

        Parameters
        ----------
        atoms : Atoms object
        
        Returns
        -------
        feature_gradient : np.ndarray
        """        
        return self.cython_module.get_featureGradient(atoms).reshape(1, len(atoms), 3, -1)

    def get_number_of_centers(self, atoms) -> int:
        """
        Get the number of centers for the given atoms object.
        Because this is a global descriptor, this is always 1.

        Parameters
        ----------
        atoms : Atoms object

        Returns
        -------
        int
            The number of centers that the descriptor will be calculated for.
            Always 1.
        """
        return 1

    @classmethod
    def from_atoms(cls, atoms, **kwargs):
        """
        Create a descriptor from an atoms object.

        Parameters
        ----------
        atoms : Atoms object
            The atoms object used to setup the descriptor.
        kwargs : dict
            Any keyword arguments to pass to the descriptor.
        """
        from agox.environments import Environment

        environment = Environment(
            template=atoms, symbols="", use_box_constraint=False, print_report=False
        )
        return cls(environment=environment, **kwargs)

    


import os

import matplotlib

matplotlib.use("Agg")
from argparse import ArgumentParser

import numpy as np
from agox import AGOX
from agox.databases import Database
from agox.environments import Environment
from agox.evaluators import LocalOptimizationEvaluator
from agox.generators.block_generators import RattleBlockGenerator
from agox.generators.MD import MDgenerator
from agox.samplers import MetropolisSampler
from ase import Atoms

# Manually set seed and database-index
seed = 42
database_index = 0

# Using argparse if e.g. using array-jobs on Slurm to do several independent searches.
# from argparse import ArgumentParser
# parser = ArgumentParser()
# parser.add_argument('-i', '--run_idx', type=int, default=0)
# args = parser.parse_args()

# seed = args.run_idx
# database_index = args.run_idx

##############################################################################
# Calculator
##############################################################################

from ase.calculators.emt import EMT

calc = EMT()

##############################################################################
# System & general settings:
##############################################################################

template = Atoms("", cell=np.eye(3) * 12)
confinement_cell = np.eye(3) * 6
confinement_corner = np.array([3, 3, 3])
environment = Environment(
    template=template,
    symbols="Au2Ni2",
    confinement_cell=confinement_cell,
    confinement_corner=confinement_corner,
)

# Database
db_path = "db{}.db".format(database_index)  # From input argument!
database = Database(filename=db_path, order=3)

##############################################################################
# Search Settings:
##############################################################################

sampler = MetropolisSampler(temperature=0.25, order=3)

generator = MDgenerator(
    **environment.get_confinement(),
    calculator=calc,
    temperature_program=[(50, 5)],  # 50 K for 5 steps    
    constraints=environment.get_constraints(),    
    environment=environment,
    sampler=sampler,    
    order=1,
)

# Do single-point relaxation after MD
evaluator = LocalOptimizationEvaluator(
    calc,
    gets={"get_key": "candidates"},
    optimizer_run_kwargs={"steps": 0},
    store_trajectory=False,
    order=2,
    constraints=environment.get_constraints(),
)

##############################################################################
# Let get the show running!
##############################################################################

agox = AGOX(generator, database, sampler, evaluator, seed=seed)

agox.run(N_iterations=10)



import numpy as np
from sklearn.metrics import mean_absolute_error, mean_squared_error, r2_score, median_absolute_error

def get_metrics(true, pred):
    """Get metrics for a regression problem.

    Args:
        true (np.ndarray): true values
        pred (np.ndarray): predicted values

    Returns:
        dict: dictionary of metrics
    """
    

    metrics = {}
    metrics['mae'] = mean_absolute_error(true, pred)
    metrics['rmse'] = np.sqrt(mean_squared_error(true, pred))
    metrics['r2'] = r2_score(true, pred)
    metrics['mdae'] = median_absolute_error(true, pred)
    return metrics




import numpy as np
from ase.constraints import FixAtoms, FixedLine, FixedPlane
from ase.io import read, write
from ase.md.langevin import Langevin
from ase.md.velocitydistribution import (MaxwellBoltzmannDistribution,
                                         Stationary, ZeroRotation)
from ase.units import fs, kB

from agox.generators.ABC_generator import GeneratorBaseClass


class MDgenerator(GeneratorBaseClass):

    """
    Does a molecular dynamics simulation on the candidate structure to propose a new candidate.

    Parameters:
    -----------
    calculator : ASE calculator
        Calculator for the MD simulation.
    thermostat : ASE thermostat
        MD program used.
    thermostat_kwargs : dict
        Settings for MD program.
    start_settings : list
        List of start settings for the MD simulation.
    start_settings_kwargs : list
        List of dictionaries with settings for start settings.
    temperature_program : list
        List of tuples with temperature and steps for MD program if temperature is modifiable during simulation.
    set_start_settings : bool
        Whether to set start settings or not.
    constraints : list
        Constraints besides fixed template and 1D/2D constraints.
    check_template : bool
        Check if template atoms moved during MD simulation.
    """

    name = "MDgenerator"

    def __init__(
        self,
        calculator,
        thermostat=Langevin,
        thermostat_kwargs={"timestep": 1.0 * fs, "temperature_K": 10, "friction": 0.05},
        start_settings=[MaxwellBoltzmannDistribution, ZeroRotation, Stationary],
        start_settings_kwargs=[{"temperature_K": 100}, {}, {}],
        temperature_program=[(500, 10), (100, 10)],
        set_start_settings=False,
        constraints=[],
        check_template=False,
        **kwargs
    ):
        super().__init__(**kwargs)

        self.calculator = calculator  # Calculator for MD simulation
        self.thermostat = thermostat  # MD program used
        self.thermostat_kwargs = thermostat_kwargs  # Settings for MD program
        self.set_start_settings = set_start_settings
        self.start_settings = start_settings
        self.start_settings_kwargs = start_settings_kwargs
        self.temperature_program = temperature_program  # (temp in kelvin, steps) for MD program if temperature is modifiable during simulation
        self.constraints = (
            constraints  # Constraints besides fixed template and 1D/2D constraints
        )

        self.check_template = (
            check_template  # Check if template atoms moved during MD simulation
        )

    def _get_candidates(self, candidate, parents, environment):
        candidate.set_calculator(self.calculator)

        self.remove_constraints(
            candidate
        )  # All constraints are removed from candidate before applying self.constraints to ensure only constraints set by user are present during MD simulation
        self.apply_constraints(candidate)
        self.molecular_dynamics(candidate)
        self.remove_constraints(
            candidate
        )  # All constraints are removed after MD simulation to not interfere with other AGOX modules

        candidate.add_meta_information("description", self.name)

        return [candidate]

    def molecular_dynamics(self, candidate):
        """Runs the molecular dynamics simulation and applies/removes constraints accordingly"""

        if self.set_start_settings:
            for setting, kwargs in zip(self.start_settings, self.start_settings_kwargs):
                setting(candidate, **kwargs)

        temperature_program = []
        for temp, steps in self.temperature_program:
            t = np.random.normal(temp, 0.1 * temp)
            temperature_program.append((t, steps))

        dyn = self.thermostat(candidate, **self.thermostat_kwargs)

        if self.check_template:
            positions_before = candidate.template.positions

        for temp, steps in temperature_program:
            dyn.set_temperature(temperature_K=temp)
            dyn.run(steps)

        if self.check_template:
            positions_after = candidate.template.positions

            if np.array_equal(positions_before, positions_after):
                self.writer("Template positions were not altered by MD simulation")
                # print('Template positions were not altered by MD simulation')
            else:
                self.writer("Template positions were altered by MD simulation")
                # print('Template positions were altered by MD simulation')

    def apply_constraints(self, candidate):
        """Applies constraints manually set and based on dimensionality of confinement cell"""

        constraints = [] + self.constraints  # Add any passed constraints immediately

        if (
            self.dimensionality == 1 or self.dimensionality == 2
        ):  # Ensures movement within 1D or 2D confinement cell
            constraints.append(self.get_dimensionality_constraints(candidate))

        candidate.set_constraint(constraints)

    def get_dimensionality_constraints(self, candidate):
        """Depending on the dimensionality this either sets fixed line or fixed plane.
        Similar to how rattle needs dimensionality specified to match with confinement cell
        """

        template = candidate.get_template()
        n_template = len(template)
        n_total = len(candidate)

        if self.dimensionality == 1:
            return FixedLine(
                indices=np.arange(n_template, n_total), direction=[1, 0, 0]
            )  # Generator assumes 1D search happens in X direction
        if self.dimensionality == 2:
            return FixedPlane(
                indices=np.arange(n_template, n_total), direction=[0, 0, 1]
            )  # Generator assumes 2D search happens in XY-plane

    def remove_constraints(self, candidate):
        """Removes all constraints from candidate"""

        candidate.set_constraint([])

    def get_number_of_parents(self, sampler):
        return 1



import numpy as np
from abc import ABC, abstractmethod

from agox.candidates import StandardCandidate
from agox.writer import header_print, pretty_print
from agox.module import Module
import functools
from copy import copy

global A
A = 0
def get_next_key():
    """
    Generates a unique always increasing key for observer methods. 

    Returns
    --------
    int: 
        Unique key. 
    """
    global A
    A += 1
    return A

class ObserverHandler:
    """
    Base-class for classes that can have attached observers. 
    """

    def __init__(self, handler_identifier, dispatch_method):
        self.observers = {}
        self.execution_sort_idx = []
        self.handler_identifier = handler_identifier
        self.set_dispatch_method(dispatch_method)

    ####################################################################################################################
    # General handling methods.
    ####################################################################################################################

    def attach_observer(self, observer_method):
        """
        Attaches an observer, saving it to the internal dict and re-evaluating 
        the execution order. 

        Parameters
        -----------
        observer_method: object
            An instance of the ObserverMethod class that will be attached. 
        """
        self.observers[observer_method.key] = observer_method
        self.evaluate_execution_order()

    def delete_observer(self, observer_method):
        """
        Removes an observer, deleting it from the internal dict and re-evaluating 
        the execution order. 

        Parameters
        -----------
        observer_method: object
            An instance of the ObserverMethod class that will be attached. 
        """
        del self.observers[observer_method.key]
        self.evaluate_execution_order()

    def evaluate_execution_order(self):
        """
        Evaluates the execution order by sorting the orders of stored 
        observers. 
        """
        keys = self.observers.keys()
        orders = [self.observers[key]['order'] for key in keys]
        self.execution_sort_idx = np.argsort(orders)

    def get_max_order(self):
        """
        Returns
        --------
        int
            The maximum order of the observers. 
        """
        return max([obs['order'] for obs in self.observers.values()])

    def get_observers_in_execution_order(self):
        """
        Returns
        --------
        List
            Observers sorted in order of execution.
        """
        observers = [obs for obs in self.observers.values()]
        sort_idx = self.execution_sort_idx
        return [observers[p]['method'] for p in sort_idx]

    def dispatch_to_observers(self, *args, **kwargs):
        """
        Dispatch to observers. 

        Only rely on the order of execution if you have specified the 'order' argument for each observer. 
        """
        for observer_method in self.get_observers_in_execution_order():
            observer_method(*args, **kwargs)

    def set_dispatch_method(self, method):
        self.dispatch_method = self.name + '.' + method.__name__

    ####################################################################################################################
    # Printing / Reporting 
    ####################################################################################################################

    def print_observers(self, include_observer=False, verbose=0, hide_log=True):
        """
        Parameters
        -----------
        include_observer: bool (default: False)
            Turns printing of observer name on. Might be useful to debug if 
            something is not working as expected. 
        verbose: int (default: 0)
            If verbose > 1 then more information is printed. 
        hide_log: bool (default: True)
            If True then the LogEntry observers are not printed. Keeps the 
            report more clean. 
        """
        from agox.tracker import TimerEntry

        order_indexs = self.execution_sort_idx
        keys = [key for key in self.observers.keys()]
        names = [obs['name'] for obs in self.observers.values()]
        methods = [obs['method'] for obs in self.observers.values()]
        orders = [obs['order'] for obs in self.observers.values()]
        class_references = [obs['class_reference'] for obs in self.observers.values()]
        
        base_string = '{}: order = {} - name = {} - method - {}'
        if include_observer:
            base_string += ' - method: {}'

        #start_string = '|'+'=' * 33 + ' Observers ' + '='*33+'|'
        header_print('Observers')
        for idx in order_indexs:
            
            if isinstance(class_references[idx], TimerEntry):
                continue

            pretty_print('  Order {} - Name: {}'.format(orders[idx], names[idx]))
            if verbose > 1:
                pretty_print('  Key: {}'.format(keys[idx]))
                pretty_print('  Method: {}'.format(methods[idx]))
                pretty_print('_'*50)

    def observer_reports(self, report_key=False, hide_log=True):
        """
        Generate observer report, which checks if the data flow is valid. 

        Parameters
        -----------
        report_key: bool
            Whether or not to print the keys used to get or set from the cache. 
        hide_log: bool
            Whether or not to print the LogEntry observers. 
        """
        from agox.tracker import TimerEntry

        dicts_out_of_order = [value for value in  self.observers.values()]
        
        header_print('Observers set/get reports')

        base_offset = '  '
        extra_offset = base_offset + '    '        
        for i in self.execution_sort_idx:

            observer_method = dicts_out_of_order[i]
            if isinstance(observer_method.class_reference, TimerEntry) and hide_log:
                continue
            # if 'Timer' in observer_method.name and hide_log:
            #     continue

            pretty_print(base_offset + observer_method.name)
            report = observer_method.report(offset=extra_offset, report_key=report_key, print_report=False, return_report=True)
            for string in report:
                pretty_print(string)
            else:
                #print(f"{dicts_out_of_order['name']} cannot report on its behavior!")
                pass

        get_set, set_set = self.get_set_match()
        pretty_print(base_offset)
        pretty_print(base_offset+'Overall:')
        pretty_print(base_offset+f'Get keys: {get_set}')
        pretty_print(base_offset+f'Set keys: {set_set}')
        pretty_print(base_offset+f'Key match: {get_set==set_set}')
        if not get_set==set_set:
            pretty_print(base_offset+'Sets do not match, this can be problematic!')
            if len(get_set) > len(set_set):
                pretty_print(base_offset+'Automatic check shows observers will attempt to get un-set item!')
                pretty_print(base_offset+'Program likely to crash!')
            if len(set_set) > len(get_set):
                pretty_print(base_offset+'Automatic check shows observers set value that is unused!')
                pretty_print(base_offset+'May cause unintended behaviour!')

            unmatched_keys = list(get_set.difference(set_set))+list(set_set.difference(get_set))
            pretty_print(base_offset+f'Umatched keys {unmatched_keys}')

    def get_set_match(self):
        """
        Check if gets and sets match.
        """
        dicts_out_of_order = [value for value in  self.observers.values()]
        all_sets = []
        all_gets = []

        for observer_method in dicts_out_of_order:
            all_sets += observer_method.sets.values()
            all_gets += observer_method.gets.values()

        all_sets = set(all_sets)
        all_gets = set(all_gets)

        return all_gets, all_sets        
            
class FinalizationHandler:
    """
    Just stores information about functions to be called when finalizaing a run. 
    """

    def __init__(self):
        self.finalization_methods = {}
        self.names = {}

    def attach_finalization(self, name, method):
        """
        Attaches finalization method.

        Parameters
        -----------
        name: str
            Human readable name of the attached method.
        method: method
            A method or function to attach. 
        """
        key = get_next_key()
        self.finalization_methods[key] = method
        self.names[key] = name
    
    def get_finalization_methods(self):
        """
        Returns
        --------
        List
            List of finalization methods.
        """
        return self.finalization_methods.values()

    def print_finalization(self, include_observer=False, verbose=0):
        """
        include_observer flag might be useful to debug if something is not working as expected. 
        """

        names = [self.names[key] for key in self.finalization_methods.keys()]
        
        base_string = '{}: order = {} - name = {} - method - {}'
        if include_observer:
            base_string += ' - method: {}'

        print('=' * 24 + ' Finalization ' + '='*24)
        for name in names:
            print('Name: {}'.format(name))
        print('='*len('=' * 25 + ' Observers ' + '='*25))
    
class Observer(Module):

    def __init__(self, gets=[dict()], sets=[dict()], order=[0], surname='', **kwargs):
        """
        Base-class for classes that act as observers. 

        Parameters
        -----------
        gets: dict
        Dict where the keys will be set as the name of attributes with the 
        value being the value of the attribute. Used to get something from the 
        iteration_cache during a run. 

        sets: dict
        Dict where the keys will be set as the name of attributes with the 
        value being the value of the attribute. Used to set something from the 
        iteration_cache during a run. 

        order: int/float
        Specifies the (relative) order of when the observer will be executed, 
        lower numbers are executed first. 

        sur_name: str
            An additional name added to classes name, can be used to distinguish 
            between instances of the same class. 
        """
        Module.__init__(self, surname=surname)

        if type(gets) == dict:
            gets = [gets]
        if type(sets) == dict:
            sets = [sets]
        if type(order) == int or type(order) == float:
            order = [order]

        combined = dict()
        for tuple_of_dicts in [gets, sets]:
            for dict_ in tuple_of_dicts:
                for key, item in dict_.items():
                    combined[key] = item
        #self.__dict__ = combined
        for key, value in combined.items():
            if key not in self.__dict__.keys():
                self.__dict__[key] = value
            else:
                raise BadOptionError('{}-set or get key has the same name as an attribute, this is not allowed'.format(key))

        self.set_keys = sum([list(set_dict.keys()) for set_dict in sets], [])
        self.set_values = sum([list(set_dict.values()) for set_dict in sets], [])
        self.get_keys = sum([list(get_dict.keys()) for get_dict in gets], [])
        self.get_values = sum([list(get_dict.values()) for get_dict in gets], [])
        self.gets = gets
        self.sets = sets

        self.order = order

        self.observer_methods = {}
        self.observer_handler_identifiers = []

        if len(kwargs) > 0:
            print('Unused key word arguments supplied to {}'.format(self.name))
            for key, value in kwargs.items():
                print(key, value)

        self.iteration_counter = None
        self.candidate_instanstiator = StandardCandidate
        
    def get_from_cache(self, key):
        """
        Attempts to get from the cache, but it is checked that the class instance 
        that tries to get has registered that it will get something with the 
        given key. 

        Parameters:
        -----------
        key: str
            The key used to index into the cache. 
        """
        assert key in self.get_values # Makes sure the module has said it wants to get with this key. 
        return self.main_get_from_cache(key)

    def add_to_cache(self, key, data, mode):
        """
        Attempts to add to the cache, but it is checked that the class instance 
        that tries to add has registered that it will add something with the 
        given key. 

        Parameters:
        -----------
        key: str
            The key used to index into the cache. 
        """
        assert key in self.set_values # Makes sure the module has said it wants to set with this key. 
        return self.main_add_to_cache(key, data, mode)

    def add_observer_method(self, method, sets, gets, order, handler_identifier='any'):
        """
        Adds an observer method that will later be attached with 'self.attach'. 

        Parameters
        -----------
        method: method
            The function/method that will be called by the observer-loop.
        sets: dict
            Dict containing the keys that the method will set in the cache. 
        gets: dict
            Dict containing the keys that the method will get from the cache. 
        """
        observer_method = ObserverMethod(self.__name__, method.__name__, method, gets, sets, order, handler_identifier)
        self.observer_methods[observer_method.key] = observer_method

    def remove_observer_method(self, observer_method):
        """
        Remove an observer_method from the internal dict of methods, if this is 
        done before calling 'attach' it will not be added as an observer. 

        Parameters
        -----------
        observer_method: object
            An instance of ObserverMethod.
        """
        key = observer_method.key
        if key in self.observer_methods.keys():
            del self.observer_methods[key]

    def update_order(self, observer_method, order):
        """
        Change the order of a method. Not really tested in practice. 

        Parameters
        -----------
        observer_method: object
            An instance of ObserverMethod.

        order: float
            New order to set.         
        """
        key = observer_method.key
        assert key in self.observer_methods.keys()
        self.observer_methods[key].order = order

    def attach(self, handler):
        """
        Attaches all ObserverMethod's as to observer-loop of the ObserverHandler 
        object 'main'
        
        Parameters
        ----------
        handler: object
            An instance of an ObserverHandler. 
        """
        for observer_method in self.observer_methods.values():
            if observer_method.handler_identifier == 'any' or observer_method.handler_identifier == handler.handler_identifier:
                handler.attach_observer(observer_method)
                self.observer_handler_identifiers.append(handler.handler_identifier)

    def reset_observer(self):
        """
        Reset the observer_methods dict, removing all observer methods.
        """
        self.observer_methods = {}

    ############################################################################
    # State
    ############################################################################

    def observer_method(func):
        @functools.wraps(func)
        def wrapper(self, state, *args, **kwargs):
            self.state_update(state)
            return func(self, state, *args, **kwargs)
        return wrapper
            
    def state_update(self, state):
        self.set_iteration_counter(state.get_iteration_counter())
    
    ############################################################################
    # Iteration counter methods:
    ############################################################################

    def get_iteration_counter(self):
        return self.iteration_counter
    
    def set_iteration_counter(self, iteration_count):
        self.iteration_counter = iteration_count

    def check_iteration_counter(self, count):
        if self.iteration_counter is None:
            return True
        return self.iteration_counter >= count

    ############################################################################
    # Checker
    ############################################################################

    def do_check(self, **kwargs):
        """
        Check if the method will run or just do nothing this iteration. 

        Returns
        -------
        bool
            True if function will run, False otherwise.w
        """
        return True

    ############################################################################
    # Candidate instanstiator
    ############################################################################

    def set_candidate_instanstiator(self, candidate_instanstiator):
        self.candidate_instanstiator = copy(candidate_instanstiator)

class ObserverMethod:

    def __init__(self, class_name, method_name, method, gets, sets, order, handler_identifier):
        """
        ObserverMethod class. Holds all information about the method that an 
        observer class attaches to the observer-loop. 

        Parameters
        ----------
        name: str
            Name of the class
        method_name: str
            Name of the method.
        method: method
            The method that is called by the observer-loop.
        sets: dict
            Dict containing the keys that the method will set in the cache. 
        gets: dict
            Dict containing the keys that the method will get from the cache. 
        order: float
            Execution order
        """
        self.class_name = class_name
        self.method_name = method_name
        self.method = method
        self.gets = gets
        self.sets = sets
        self.order = order

        self.name = self.class_name + '.' + self.method_name
        self.class_reference = method.__self__
        #self.key = method.__hash__() # Old
        self.key = get_next_key()
        self.handler_identifier = handler_identifier

    def __getitem__(self, key):
        return self.__dict__[key]

    def report(self, offset='', report_key=False, return_report=False, print_report=True):
        """
        Generate report. Used by ObserverHandler.observer_reports

        Parameters
        ----------
        offset: str
        report_key: bool
            Print the key or not. 
        return_report: bool
            Whether to return the report or not.
        print_report: bool
            Whether to print hte print or not. 

        Returns
        -------
        str or None:
            If report_key = True then it returns a string otherwise None is r
            eturned. 
        """
        report = []

        for key in self.gets.keys():
            value = self.gets[key]
            out = offset + f"Gets '{value}'" 
            if report_key:
                out += f" using key attribute '{key}'"
            report.append(out)

        for key in self.sets.keys():
            value = self.sets[key]
            out = offset + f"Sets '{value}'"
            if report_key:
                out += f" using key attribute '{key}'"
            report.append(out)
        
        if len(self.gets) == 0 and len(self.sets) == 0:
            out = offset+'Doesnt set/get anything'
            report.append(out)

        if print_report:
            for string in report:
                print(string)
        if return_report:
            return report

if __name__ == '__main__':

    # class TestObserverObject:

    #     def __init__(self, name):
    #         self.name = name

    #     def test_observer_method(self):
    #         #print('Name {} - I am method {} on {}'.format(id(self.test_observer_method), id(self)))
    #         print('My name is {}'.format(self.name))

    # handler = ObserverHandler()

    # A = TestObserverObject('A')
    # B = TestObserverObject('B')

    # handler.add_observer(A.test_observer_method,'A', order=0)
    # handler.add_observer(B.test_observer_method,'B', order=0)
    # #handler.add_observer(B.test_observer_method,'B', order=-1)

    # for method in handler.get_observers_in_execution_order():
    #     method()

    #for key in handler.observers.keys():
    #    print(key, handler.observers[key])        

    # print(dir(B.test_observer_method))
    # print(B.test_observer_method.__self__)

    observer = Observer(get_key='candidates', set_key='candidates', bad_key='bad')
    
    print(observer.__dict__)

    observer.report()


        
    


"""
Collectors are used to collect multiple candidates generated by generators. 
This is a task that can be parallelized over the number of generated candidates, 
and this is done using parallelization from Ray. 
"""
from agox.collectors.standard import StandardCollector
from agox.collectors.ray_collector import ParallelCollector
from agox.collectors.parallel_tempering import ParallelTemperingCollector

__all__ = ['StandardCollector', 'ParallelCollector', 'ParallelTemperingCollector']



import numpy as np
from agox.postprocessors.ABC_postprocess import PostprocessBaseClass
from ase.calculators.calculator import all_properties
from ase.calculators.singlepoint import SinglePointCalculator
from ase.constraints import FixAtoms
from ase.optimize import BFGS

class RelaxPostprocess(PostprocessBaseClass):

    """
    Relaxes candidates using a local optimization algorithm.

    Parameters
    ----------
    model : agox.models.ABC_model
        Model to use for relaxation.
    optimizer : ase.optimize.Optimizer
        Optimizer to use for relaxation.
    optimizer_run_kwargs : dict
        Keyword arguments to pass to the optimizer.run method.
    optimizer_kwargs : dict
        Keyword arguments to pass to the optimizer constructor.
    fix_template : bool
        If True, the template atoms are fixed during relaxation.
    constraints : list
        List of constraints to apply during relaxation.
    start_relax : int
        The iteration to start relaxing candidates.    
    """

    name = 'PostprocessRelax'
    
    def __init__(self, model=None, optimizer=None, optimizer_kwargs={'logfile':None},
        optimizer_run_kwargs={'fmax':0.05, 'steps':200}, fix_template=True, 
        constraints=[], start_relax=1, **kwargs):
        super().__init__(**kwargs)
        self.optimizer = BFGS if optimizer is None else optimizer
        self.optimizer_kwargs = optimizer_kwargs
        self.optimizer_run_kwargs = optimizer_run_kwargs
        self.start_relax = start_relax
        self.model = model

        # Constraints:
        self.constraints = constraints
        self.fix_template = fix_template
    
    def postprocess(self, candidate):
        initial_candidate = candidate.copy()
        candidate.set_calculator(self.model)
        self.apply_constraints(candidate)
        optimizer = self.optimizer(candidate, **self.optimizer_kwargs)

        try: 
            optimizer.run(**self.optimizer_run_kwargs)
        except Exception as e:
            print('Relaxation failed with exception: {}'.format(e))
            return initial_candidate
        
        candidate.add_meta_information('relaxation_steps', optimizer.get_number_of_steps())

        results = {prop: val for prop, val in candidate.calc.results.items() if prop in all_properties}
        candidate.calc = SinglePointCalculator(candidate, **results)

        print(f'Relaxed for {optimizer.get_number_of_steps()} steps')

        self.remove_constraints(candidate)
        return candidate

    def do_check(self, **kwargs):
        return self.check_iteration_counter(self.start_relax) * self.model.ready_state

    ####################################################################################################################
    # Constraints
    ####################################################################################################################

    def apply_constraints(self, candidate):
        constraints = [] + self.constraints
        if self.fix_template:
            constraints.append(self.get_template_constraint(candidate))

        for constraint in constraints:
            if hasattr(constraint, 'reset'):
                constraint.reset()

        candidate.set_constraint(constraints)

    def remove_constraints(self, candidate):
        candidate.set_constraint([])

    def get_template_constraint(self, candidate):
        return FixAtoms(indices=np.arange(len(candidate.template)))



from typing import List

import numpy as np
from ase import Atoms

from agox.utils.filters.ABC_filter import FilterBaseClass


class RandomFilter(FilterBaseClass):
    """
    Filter that removes all structures.

    """

    name = "RandomFilter"

    def __init__(self, N, **kwargs):
        """
        Parameters
        ----------
        """
        super().__init__(**kwargs)
        self.N = N

    def _filter(self, atoms: List[Atoms]) -> np.ndarray:
        if len(atoms) < self.N:
            return np.arange(len(atoms))
        else:
            return np.random.choice(np.arange(len(atoms)), self.N, replace=False)



from ase.io import read, write
from agox.environments.environment import Environment

from agox.generators import SymmetryGenerator, SymmetryRattleGenerator, SymmetryPermutationGenerator
import numpy as np
from ase import Atoms
from agox.samplers.fixed_sampler import FixedSampler

from agox.candidates import CandidateBaseClass

def environment_factory():
    template = Atoms('', cell=np.eye(3)*22, pbc=False)
    confinement_corner = np.array([5.5, 5.5, 5.5])
    confinement_cell = np.eye(3)*11
    environment = Environment(template=template, symbols='B12N12',
        confinement_corner=confinement_corner,
        confinement_cell=confinement_cell)
    return environment

def generator_factory(environment):
    generator = SymmetryGenerator(sym_type='cluster', force_group='D6', 
                            may_nucleate_at_several_places=True,
                            **environment.get_confinement())
    return generator

def rattle_generator_factory(environment):
    rattle_generator = SymmetryRattleGenerator(**environment.get_confinement(), 
                                               let_lose_sym=1.0)
    return rattle_generator

def permutation_generator_factory(environment):
    permutation_generator = SymmetryPermutationGenerator(**environment.get_confinement())
    return permutation_generator

def test_symmetry_generator():
    np.random.seed(1)
    environment = environment_factory()
    generator = generator_factory(environment)
    candidate = generator(None, environment)[0]
    assert isinstance(candidate, CandidateBaseClass)

def test_symmetry_rattle_generator():
    np.random.seed(1)
    environment = environment_factory()
    generator = generator_factory(environment)
    rattle_generator = rattle_generator_factory(environment)
    symmetry_candidate = generator(None, environment)[0]
    sampler = FixedSampler(symmetry_candidate)
    candidate = rattle_generator(sampler, environment)[0]
    assert isinstance(candidate, CandidateBaseClass)

def test_symmetry_permutation_generator():
    np.random.seed(1)
    environment = environment_factory()
    generator = generator_factory(environment)
    permutation_generator = permutation_generator_factory(environment)
    symmetry_candidate = generator(None, environment)[0]
    sampler = FixedSampler(symmetry_candidate)
    candidate = permutation_generator(sampler, environment)[0]
    assert isinstance(candidate, CandidateBaseClass)



from .cli_analysis import add_analysis_parser
from .cli_convert import add_convert_parser
from .cli_notebook import add_notebook_parser
from .cli_graph_sorting import add_graph_sorting_parser

parsers = [add_analysis_parser, add_convert_parser, add_notebook_parser, add_graph_sorting_parser]


from agox.test.run_tests.run_utils import agox_test_run
import pytest

# @pytest.mark.skip(reason='Not reproducible on CI at the moment.')
def test_rss_cluster_symmetry(tmp_path, cmd_options):
    mode = 'rss_cluster_symmetry' # This determines the script that is imported.
    agox_test_run(mode, tmp_path, cmd_options, full_compare=False)


import numpy as np
from agox.databases.database_concurrent import ConcurrentDatabase


class OrderedDatabase(ConcurrentDatabase):
        
    def restore_to_memory(self):
        strucs = self.get_all_structures_data()
        if self.store_meta_information:
            all_meta_info = self.read_all_key_value_pairs()
        else:
            all_meta_info = {}

        all_candidates = []
        for struc in strucs:            
            all_candidates.append(self.db_to_candidate(struc, meta_dict=all_meta_info.get(struc['id'], None)))

        iteration = self.get_iteration_counter()
        
        new_candidates = [c for c in all_candidates if c.get_meta_information('iteration') > iteration - self.sync_frequency]

        old_candidates = [c for c in self.candidates if c.get_meta_information('iteration') <= iteration - self.sync_frequency]

        worker_numbers = np.array([c.get_meta_information('worker_number') != self.worker_number for c in new_candidates],
                                  dtype=int)
        iterations = np.array([c.get_meta_information('iteration') for c in new_candidates])
        relax_indices = np.array([c.get_meta_information('relax_index') for c in new_candidates])
        indices = np.lexsort((iterations, worker_numbers))

        self.candidates =  old_candidates + [new_candidates[i] for i in indices]

        if self.verbose > 1:
            for c in self.candidates:
                self.writer(f"worker number: {c.get_meta_information('worker_number')}, iteration: {c.get_meta_information('iteration')}, relax_index: {c.get_meta_information('relax_index')}")        



import pytest
import numpy as np
from agox.models.GPR import GPR

from agox.models.GPR.kernels import RBF, Constant as C, Noise
from agox.models.descriptors import JaxFingerprint


from agox.test.model_tests.model_utils import model_tester
from agox.test.test_utils import get_test_data, get_test_environment, load_expected_data, save_expected_data, label_dict_list
from agox.test.test_utils import test_data_dicts, get_name, check_file_is_deleted

kernel = C() * RBF() + Noise(0.01)
model_class = GPR
model_maker = GPR
model_base_args = []
model_base_kwargs = {'use_ray':False}

model_update_kwargs = [
    {"kernel": kernel}
]

model_update_kwargs = label_dict_list(model_update_kwargs)

@pytest.fixture(params=model_update_kwargs)
def update_kwargs(request):
    return request.param

@pytest.mark.ray
@pytest.mark.parametrize('test_data_dict', test_data_dicts)
def test_model(test_data_dict, update_kwargs, cmd_options):

    create_mode = cmd_options['create_mode']
    test_mode = not create_mode
    tolerance = cmd_options['tolerance']

    update_kwargs, parameter_index  = update_kwargs

    path = test_data_dict['path']
    remove = test_data_dict['remove']
    dataset_name = test_data_dict['name']
    parameter_index = 0

    environment = get_test_environment(path, remove)
    data = get_test_data(path, environment)[0:10]

    # Update the kwargs for the model. 
    descriptor = JaxFingerprint.from_atoms(data[0], r_cut_radial=4.0, r_cut_angular=4.0)
    update_kwargs['descriptor'] = descriptor

    model_kwargs = model_base_kwargs.copy()
    model_kwargs.update(update_kwargs)    

    # Where to load/save data. 
    subfolder = 'model_tests/'
    module_name = 'GlobalGPRJax'
    name = get_name(module_name, subfolder, dataset_name, parameter_index)
    if test_mode:
        expected_data = load_expected_data(name) 
    else:
        expected_data = None

    output = model_tester(model_maker, model_base_args, model_kwargs, data, 
        test_mode=test_mode, expected_data=expected_data, tolerance=tolerance, 
        force_tolerance=dict(rtol=1e-2, atol=1e-3))

    if not test_mode:
        check_file_is_deleted(name)
        save_expected_data(name, output)








import numpy as np
from scipy.stats import norm
from agox.acquisitors import AcquisitorBaseClass
from agox.observer import Observer
from agox.writer import agox_writer

class ExpectedImprovementAcquisitor(AcquisitorBaseClass):
    
    name = 'EIAcquisitor'
    
    def __init__(self, model, database, xi=0.01, **kwargs):
        super().__init__(maximize=True, **kwargs)
        self.model = model
        self.xi = xi
        self.lowest_energy = 10E10
        self.add_observer_method(self.update, sets={}, gets={}, order=0, handler_identifier='database')
        self.attach(database)

    def calculate_acquisition_function(self, candidates):
        fitness = np.zeros(len(candidates))
        for i, candidate in enumerate(candidates):
            E, sigma = self.model.predict_energy_and_uncertainty(candidate)
            fitness[i] = self.acquisition_function(E, sigma)
            candidate.add_meta_information('model_energy', E)
            candidate.add_meta_information('uncertainty', sigma)
        return fitness
    
    def acquisition_function(self, E, sigma):
        mu = E
        mu_opt = self.lowest_energy
        with np.errstate(divide='warn'):
            imp = mu_opt - mu - self.xi
            Z = imp / sigma 
            ei = imp * norm.cdf(Z) + sigma * norm.pdf(Z)
        return ei
    
    def print_information(self, candidates, acquisition_values):
        if self.do_check():
            for i, candidate in enumerate(candidates):
                fitness = acquisition_values[i]
                E = candidate.get_meta_information('model_energy')
                sigma = candidate.get_meta_information('uncertainty')
                self.writer(f'Candidate {i}: E = {E:8.3f}, s = {sigma:8.3f}, EI = {fitness:8.6f}')

    @agox_writer
    @Observer.observer_method
    def update(self, database, state):
        self.lowest_energy = database.get_best_energy()
        self.writer('Lowest energy: {}'.format(self.lowest_energy))

    def do_check(self, **kwargs):
        return self.model.ready_state




import pytest 
from agox.utils.batch_analysis import Analysis
from agox.test.test_utils import TemporaryFolder, test_folder_path
import os
import shutil

import matplotlib.pyplot as plt
import glob


database_paths = [
    [os.path.join(test_folder_path, 'datasets/databases/mos2_databases/')]
]

@pytest.fixture(params=database_paths)
def database_paths(request):
    return request.param

def test_analysis(tmp_path, database_paths):

    absolute_paths = []
    for database_path in database_paths:
        path = os.path.join(tmp_path, 'dataset')
        shutil.copytree(database_path, path)
        absolute_paths.append(path)

    with TemporaryFolder(tmp_path):

        for force_reload in [True, False]:
            analysis = Analysis()
            for path in absolute_paths:
                analysis.add_directory(path, force_reload=False)
            analysis.compile_information()
            analysis.calculate_CDF()

            fig, ax = plt.subplots()
            analysis.plot_CDF(ax)
            analysis.plot_histogram(ax)
            analysis.plot_energy(ax)



from typing import Any, Dict, Optional, Union

import numpy as np
from ase.cell import Cell
from matplotlib.axes import Axes
from matplotlib.collections import PolyCollection
from numpy.typing import NDArray

from agox.utils.plot.utils import plane_to_indices


def plot_cell(ax: Axes,
              cell: Union[Cell, NDArray],
              *,
              plane: str = 'xy+',
              offset: Optional[NDArray] = None,
              collection_kwargs: Dict[str, Any] = None):
    """Plot a 2D projection of a `Cell` object into an `Axes` object.

    Parameters
    ----------
    ax : Axes
        Axes to plot the cell in.
    cell : Union[Cell, NDArray]
        Cell to plot.
    plane : str, optional
        Plane to project the cell on. See `plane_to_indices` for information on
        valid plane strings, by default 'xy+'.
    offset : Optional[NDArray], optional
        Cartesian coordinates to apply as offset to the base of the cell, by
        default None (no offset).
    collection_kwargs : Dict[str, Any], optional
        Keyword arguments additionally forwarded to the `PolyCollection`
        constructor for the cell, by default None.

    Raises
    ------
    ValueError
        Raised if `offset` or `cell` has an incorrect shape.
    """

    default_collection_kwargs = dict(
        edgecolors='k',
        facecolors='none',
        linestyles='dotted'
    )

    if offset is None:
        offset = np.zeros(3)
    else:
        offset = np.asarray(offset)

    if offset.shape != (3,):
        raise ValueError(f'`offset` must have shape (3,), input has {offset.shape}')

    if isinstance(cell, Cell):
        cell = cell.complete()

    if cell.shape != (3, 3):
        raise ValueError(f'`cell` must have shape (3, 3), input has {cell.shape}')

    if collection_kwargs is None:
        collection_kwargs = {}

    collection_kwargs = {**default_collection_kwargs, **collection_kwargs}

    d0, d1, _, _ = plane_to_indices(plane)

    verts_uc = np.array([
        [[0, 0, 0], [1, 0, 0], [1, 1, 0], [0, 1, 0]],  # xy, z = 0
        [[0, 0, 1], [1, 0, 1], [1, 1, 1], [0, 1, 1]],  # xy, z = 1
        [[0, 0, 0], [1, 0, 0], [1, 0, 1], [0, 0, 1]],  # xz, y = 0
        [[0, 1, 0], [1, 1, 0], [1, 1, 1], [0, 1, 1]],  # xz, y = 1
        [[0, 0, 0], [0, 1, 0], [0, 1, 1], [0, 0, 1]],  # yz, x = 0
        [[1, 0, 0], [1, 1, 0], [1, 1, 1], [1, 0, 1]]   # yz, x = 1
    ])
    verts_3d = verts_uc @ cell + offset
    verts_2d = verts_3d[:, :, [d0, d1]]

    p = PolyCollection(verts_2d, **collection_kwargs)
    ax.add_collection(p)
    ax.autoscale_view()



import pickle
import numpy as np 
from time import sleep
from ase import Atoms
import sqlite3
import os
import sys
import os.path

from .database_utilities import *
from .ABC_database import DatabaseBaseClass
from copy import deepcopy
from collections import OrderedDict

# Should match init_statements!!
from agox.candidates import StandardCandidate

class Database(DatabaseBaseClass):
    
    """ Database module """

    init_statements = ["""create table structures (
    id integer primary key autoincrement,
    ctime real,
    positions blob,
    energy real,
    type blob,
    cell blob,
    forces blob, 
    pbc blob,
    template_indices blob,
    iteration int,
    cache_key text
    )""", 

    """CREATE TABLE text_key_values (
    key TEXT,
    value TEXT,
    id INTEGER,
    FOREIGN KEY (id) REFERENCES systems(id))""",

    """CREATE TABLE float_key_values (
    key TEXT,
    value REAL,
    id INTEGER,
    FOREIGN KEY (id) REFERENCES systems(id))""", 

    """CREATE TABLE int_key_values (
    key TEXT,
    value INTEGER,
    id INTEGER,
    FOREIGN KEY (id) REFERENCES systems(id))""", 
    
    """CREATE TABLE boolean_key_values (
    key TEXT,
    value INTEGER,
    id INTEGER,
    FOREIGN KEY (id) REFERENCES systems(id))""", 
    
    """CREATE TABLE other_key_values (
    key TEXT,
    value BLOB,
    id INTEGER,
    FOREIGN KEY (id) REFERENCES systems(id))"""]

    #unpack_functions = [nothing, nothing, deblob, nothing, deblob, deblob, deblob, deblob, deblob, nothing]
    #pack_functions = [nothing, nothing, blob, nothing, blob, blob, blob, blob, nothing]
    
    # Pack: Positions, energy, type, cell, forces, pbc, template_indices, iteration
    # Unpack: ID, time, -//-
    pack_functions = [blob, nothing, blob, blob, blob, blob, blob, nothing, nothing]
    unpack_functions = [nothing, nothing, deblob, nothing, deblob, deblob, deblob, deblob, deblob, nothing, nothing]

    name = 'Database'

    def __init__(self, filename='db.db', initialize=False, verbose=False, write_interval=1, call_initialize=True, 
                store_meta_information=True, **kwargs):
        super().__init__(**kwargs)

        # File-based stuff:
        self.filename = filename
        if initialize and os.path.exists(filename):
            os.remove(filename)
        self.con = sqlite3.connect(filename, timeout=600)
        self.write_interval = write_interval

        # Important that this matches the init_statements list. 
        self.storage_keys = ['positions', 'energy', 'type', 'cell', 'forces', 'pbc', 'template_indices', 'iteration', 'cache_key']

        # Memory-based stuff:        
        self.candidate_instanstiator = StandardCandidate
        self.candidate_energies = []
        
        # Whether or not to save and retrieve meta information from the database.
        self.store_meta_information = store_meta_information
        self.meta_dict_list = []

        if call_initialize:
            self._initialize()

    ####################################################################################################################
    # Memory-based methods:
    ####################################################################################################################

    def store_candidate(self, candidate, accepted=True, write=True, dispatch=True):
        # Needs some way of handling a dummy candidate, probably boolean argument.
        if accepted:
            self.candidates.append(candidate)
            self.candidate_energies.append(candidate.get_potential_energy())
        if write:
            self.write(candidate)
        # if dispatch:
        #     self.dispatch_to_observers(self, state)

    def get_all_candidates(self, **kwargs):
        all_candidates = []
        for candidate in self.candidates:
            all_candidates.append(candidate)
        return all_candidates

    def get_most_recent_candidate(self):
        if len(self.candidates) > 0:
            candidate = deepcopy(self.candidates[-1])
        else:
            candidate = None
        return candidate

    def get_recent_candidates(self, number):
        return [deepcopy(candidate) for candidate in self.candidates[-number:]]

    def get_best_energy(self):
        return np.min(self.candidate_energies)
        
    ####################################################################################################################
    # File-based methods:
    ####################################################################################################################

    def _init_storage(self):
        self.positions = []
        self.atom_numbers = []
        self.energies = []
        self.cells = []
        self.forces = []
        self.pbc = []
        self.template_indices = []

        self.storage_dict = OrderedDict()
        for key in self.storage_keys:
            self.storage_dict[key] = []
        
        self.number_of_rows = 0
        self.meta_dict_list = []
        
    def _initialize(self):
        # Check if structures are in the database, otherwise initialize tables
        cur = self.con.execute(
            'select count(*) from sqlite_master where name="structures"')
        try:
            if cur.fetchone()[0] == 0:
                for statement in self.init_statements:
                    self.con.execute(statement)
                self.con.commit()
        except sqlite3.OperationalError:
            pass
        self._init_storage()
        self.con.row_factory = sqlite3.Row

    def store_information(self, candidate):
        if candidate is not None:            
            candidate.print_properties(self.writer,self.get_iteration_counter())
            e = candidate.calc.results.get('energy', 0)
            f = candidate.calc.results.get('forces', np.zeros(candidate.positions.shape))
            
            self.storage_dict['energy'].append(e)
            self.storage_dict['forces'].append(f)
            self.storage_dict['positions'].append(np.array(candidate.positions, dtype=np.float64))
            self.storage_dict['type'].append(np.array(candidate.numbers, dtype=np.float64))
            self.storage_dict['cell'].append(np.array(candidate.cell, dtype=np.float64))
            self.storage_dict['pbc'].append(np.array(candidate.pbc.astype(int), dtype=np.float64))
            self.storage_dict['template_indices'].append(np.array(candidate.get_template_indices(), dtype=np.float64))
            self.storage_dict['iteration'].append(self.get_iteration_counter())
            self.storage_dict['cache_key'].append(candidate.cache_key)
            self.number_of_rows += 1

            if self.store_meta_information:
                self.meta_dict_list.append(candidate.meta_information)

    def get_row_to_write(self, index):
        row = [now()]
        for (key, value), func in zip(self.storage_dict.items(), self.pack_functions):
            row.append(func(value[index]))

        return tuple(row)

    def get_last_id(self, cur):
        cur.execute('SELECT seq FROM sqlite_sequence WHERE name="structures"')
        result = cur.fetchone()
        if result is not None:
            id = result[0]
            return id
        else:
            return 0

    def write(self, candidate=None, force_write=False, already_stored=False):

        if not already_stored: 
            self.store_information(candidate)

        if self.number_of_rows == self.write_interval or force_write:
            try:
                for row_index in range(self.number_of_rows):
                    cur = self.con.cursor()
                    row = self.get_row_to_write(row_index)
                    q = 'NULL,' + ', '.join('?' * len(row))

                    cur.execute('insert into structures values ({})'.format(q), row)
                    
                    self.last_used_id = self.get_last_id(cur)

                    if self.store_meta_information:
                        cur = self.con.cursor()
                        id = self.last_used_id
                        self.write_key_value_pairs(cur, self.meta_dict_list[row_index], id=id)
                    self.con.commit()                    

            except sqlite3.OperationalError as error:
                print('Encountered error: {}'.format(error))
                sleep(1)
                self.write(candidate=candidate, force_write=force_write, already_stored=True)
            self._init_storage()          

    def db_to_atoms(self, structure):
        """

        Converts a database representation (dictionary) of a structure to an ASE atoms object

        Parameters
        ----------
        structure :  database representation of a structure

        Returns
        -------
        struc : ASE Atoms object

        """
        e = structure['energy']
        f = structure.get('forces', 0).reshape(-1, 3)
        pos = structure['positions']
        num = structure['type']
        cell = structure['cell']
        pbc = structure.get('pbc', None)

        atoms = Atoms(symbols = num,
                    positions = pos,
                    cell = cell, 
                    pbc=pbc)        

        calc = SinglePointCalculator(atoms, energy=e, forces=f)
        atoms.set_calculator(calc)
        return atoms

    def db_to_candidate(self, structure, meta_dict=None):
        e = structure['energy']
        f = structure.get('forces', 0).reshape(-1, 3)
        pos = structure['positions']
        num = structure['type']
        cell = structure['cell']
        pbc = structure.get('pbc', None)
        template_indices = structure.get('template_indices', None)
        cache_key = structure.get('cache_key', None)
        
        if hasattr(self, 'template') and template_indices is None:
            template = self.template
        else:
            template = None            

        candidate = self.candidate_instanstiator(symbols = num, positions = pos, cell = cell, pbc=pbc, template=template, template_indices=template_indices)        
        candidate.cache_key = cache_key
        calc = SinglePointCalculator(candidate, energy=e, forces=f)
        candidate.set_calculator(calc)

        if self.store_meta_information:
            if meta_dict is None:
                restored_dict = self.read_key_value_pairs(id=structure['id'])
                for key, value in restored_dict.items():
                    candidate.add_meta_information(key, value)
            else:
                candidate.meta_information.update(meta_dict)

        candidate.add_meta_information('iteration', structure.get('iteration', 0))
        candidate.add_meta_information('database_id', structure['id'])

        return candidate

    def get_data_from_row(self, row):
        d = {}
        for key, value, func in zip(row.keys(), row, self.unpack_functions):
            d[key] = func(value)

        d['positions'] = d['positions'].reshape(-1, 3)
        d['cell'] = d['cell'].reshape(3, 3)

        return d

    def get_all_structures_data(self):
        cursor = self.con.execute("SELECT * from structures")
        structures = []
        for row in cursor.fetchall():
            d = self.get_data_from_row(row)
            structures.append(d)
        return structures

    def get_structure_data(self, i):
        t = (str(int(i)),)
        cursor = self.con.execute('SELECT * from structures WHERE id=?',t)
        row = cursor.fetchone()
        d = self.get_data_from_row(row)    
        return d        

    def get_all_energies(self):
        cursor = self.con.execute("SELECT energy from structures")
        energies = []
        for row in cursor.fetchall():
            energies.append(row['energy'])
        return np.array(energies)

    def get_best_structure(self):
        structure_data = self.get_all_structures_data()
        energies = [c['energy'] for c in structure_data]
        idx = energies.index(min(energies))
        best_struc = self.db_to_atoms(structure_data[idx])        
        return best_struc
        
    def get_structure(self, index):
        cand = self.get_structure_data(index)
        struc = self.db_to_atoms(cand)

        return struc

    ####################################################################################################################
    # Meta-information related.
    ####################################################################################################################

    def restore_to_memory(self):
        strucs = self.get_all_structures_data()
        if self.store_meta_information:
            all_meta_info = self.read_all_key_value_pairs()
        else:
            all_meta_info = {}

        candidates = []
        for struc in strucs:            
            candidates.append(self.db_to_candidate(struc, meta_dict=all_meta_info.get(struc['id'], None)))
        self.candidates = candidates
        self.candidate_energies = [atoms.get_potential_energy() for atoms in candidates]

    def restore_to_trajectory(self):
        strucs = self.get_all_structures_data()
        atoms_objects = []
        for structure in strucs:
            atoms_objects.append(self.db_to_atoms(structure))
        return atoms_objects

    ####################################################################################################################
    # Meta-information related.
    ####################################################################################################################

    def write_key_value_pairs(self, cur, dictionary, id=0):            
        text_pairs = []
        float_pairs = []
        int_pairs = []
        bool_pairs = []
        other_pairs = []

        for key, value in dictionary.items():
            if type(value) == str:
                text_pairs += [(key, value, id)]
            elif type(value) == float:
                float_pairs += [(key, value, id)]
            elif type(value) == int:
                int_pairs += [(key, value, id)]
            elif type(value) == bool:
                bool_pairs += [(key, value, id)]
            else:
                value_converted =  pickle.dumps(value, pickle.HIGHEST_PROTOCOL)
                other_pairs += [(key, value_converted, id)]
        
        cur.executemany('INSERT INTO text_key_values VALUES (?, ?, ?)', text_pairs)
        cur.executemany('INSERT INTO float_key_values VALUES (?, ?, ?)', float_pairs)
        cur.executemany('INSERT INTO int_key_values VALUES (?, ?, ?)', int_pairs)
        cur.executemany('INSERT INTO boolean_key_values VALUES (?, ?, ?)', bool_pairs)        
        cur.executemany('INSERT INTO other_key_values VALUES (?, ?, ?)', other_pairs)

    def read_key_value_pairs(self, id=0):
        dict_recreation = {}
        tables = ['text_key_values', 'float_key_values', 'int_key_values', 'boolean_key_values', 'other_key_values']
        functions = [nothing, nothing, nothing, bool, pickle.loads]

        cur = self.con.cursor()
        for table, func in zip(tables, functions):
            cur.execute(f'SELECT * FROM {table} WHERE id=?', (id,))
            A = cur.fetchall()
            for key, value, _ in A:
                try:
                    converted_value = func(value)
                except:
                    converted_value = deblob(value)

                dict_recreation[key] = converted_value

        return dict_recreation

    def read_all_key_value_pairs(self):
        cursor = self.con.cursor()

        tables = ['text_key_values', 'float_key_values', 'int_key_values', 'boolean_key_values', 'other_key_values']
        functions = [nothing, nothing, nothing, bool, pickle.loads]

        dict_of_dicts = {} # Uses the id as the key to the meta information dict for each candidate. (Id from structures table).

        for table, func in zip(tables, functions):
            cursor.execute(f'SELECT * FROM {table}')
            rows = cursor.fetchall()

            for key, value, id in rows:
                try:
                    converted_value = func(value)
                except:
                    converted_value = deblob(value)

                if id in dict_of_dicts.keys():
                    dict_of_dicts[id][key] = converted_value
                else:
                    dict_of_dicts[id] = {key:converted_value}

        return dict_of_dicts

    def update_meta_information(self, meta_dict, database_index, memory_index):
        """

        Update meta-information of the candidate at database_index 'id' in the structures
        table. 

        Parameters
        ----------
        meta_dict : dict
            Dict containing the information to update - if keys are present that 
            are already present in the database they are overwritten with the new
            values and if new keys are present they are added to the database. 
        index : int
            The index (id) of the structure in the structures table. 
        """
        if database_index is not None:
            cur = self.con.cursor()
            self.write_key_value_pairs(cur, meta_dict, id=database_index)
            self.con.commit()
        if memory_index is not None:
            self.candidates[memory_index].meta_information.update(meta_dict)



from .ABC_sampler import SamplerBaseClass
import numpy as np
from ase.io import write
from ase.data import covalent_radii

class SpectralGraphSampler(SamplerBaseClass):

    """
    Spectral Graph Sampler for selecting candidates based on a spectral graph criterion.

    Calculates the eigenvalues of the adjacency matrix of the covalent bonds of the structure.
    The eigenvalues are then used as a fingerprint for the structure.

    Parameters
    ----------
    model_calculator : agox.models.model.ABC_model.ModelBaseClass
        Model to use for the similarity criterion.
    indices : list
        List of indices of the atoms to use in the adjacency matrix.
    sample_size : int
        The number of sample members, or population size.
    covalent_bond_scale_factor : float
        The scale factor for the covalent bond distance.
    """

    name = 'SpectralGraphSampler'

    def __init__(self, model_calculator, indices=None, debug=False, 
        sample_size=10, covalent_bond_scale_factor=1.3, **kwargs):
        super().__init__(**kwargs)
        self.indices = indices
        self.sample_size = sample_size
        self.covalent_bond_scale_factor = covalent_bond_scale_factor
        self.debug = debug
        self.model_calculator = model_calculator

    def setup(self, all_finished_structures):
        if self.debug:
            write('all_finished_structures_{:06d}.traj'.format(self.get_iteration_counter()),all_finished_structures)
        if self.verbose:
            print('len(all_finished_structures):',len(all_finished_structures))
        if len(all_finished_structures) == 0:
            return

        self.template = all_finished_structures[0].template.copy()
        all_finished_structures.sort(key=lambda x: x.get_potential_energy())

        sample_indices = []
        self.sample_eigen_strings = []
        for i,t in enumerate(all_finished_structures):
            s = self.get_eigen_value_string(t,allow_read_from_meta_information=True)
            if s not in self.sample_eigen_strings:
                sample_indices.append(i)
                self.sample_eigen_strings.append(s)
                print('SAMPLE index',i,s)
            if len(sample_indices) >= self.sample_size:
                break
        
        self.sample = [all_finished_structures[i] for i in sample_indices]
        if self.verbose:
            write('sample_{:06d}.traj'.format(self.get_iteration_counter()),self.sample)
        sample_energies = [t.get_potential_energy() for t in self.sample]
        print('SAMPLE_DFT_ENERGY  ', '[',','.join(['{:8.3f}'.format(e) for e in sample_energies]),']')

        if hasattr(self.model_calculator,'ever_trained') and self.model_calculator.ever_trained:
            for s in self.sample:
                t = s.copy()
                t.set_calculator(self.model_calculator)
                E = t.get_potential_energy()
                sigma = self.model_calculator.get_property('uncertainty')
                s.add_meta_information('model_energy',E)
                s.add_meta_information('uncertainty',sigma)
            print('SAMPLE_MODEL_ENERGY', \
                  '[',','.join(['{:8.3f}'.format(t.get_meta_information('model_energy')) for t in self.sample]),']')
            print('SAMPLE_MODEL_SIGMA', \
                  '[',','.join(['{:8.3f}'.format(t.get_meta_information('uncertainty')) for t in self.sample]),']')

    def assign_to_closest_sample(self,candidate_object):

        if len(self.sample) == 0:
            return None
        
        # find out what cluster we belong to
        s = self.get_eigen_value_string(candidate_object)

        if s in self.sample_eigen_strings:
            i = self.sample_eigen_strings.index(s)
        else:
            i = len(self.sample) - 1

        cluster_center = self.sample[i]
        cluster_center.add_meta_information('index_in_sampler_list_used_for_printing_purposes_only',i)
        return cluster_center

    def adjust_sample_cluster_distance(self, chosen_candidate, closest_sample):
        # See if this structure had a lower energy than the lowest energy in the cluster it belongs to
        chosen_candidate_energy = chosen_candidate.get_potential_energy()
        closest_sample_energy = closest_sample.get_potential_energy()
        d_min_index = closest_sample.get_meta_information('index_in_sampler_list_used_for_printing_purposes_only')
        print('CLUST_RES {:06d}'.format(self.get_iteration_counter()),
              '[',','.join(['{:8.3f}'.format(t.get_potential_energy()) for t in self.sample]),
              '] {:8.3f} {:8.3f}'.format(closest_sample_energy,chosen_candidate_energy),
              'NEW {:d} {}'.format(d_min_index,chosen_candidate.get_meta_information('description')) \
              if closest_sample_energy > chosen_candidate_energy else '')

    def get_distances(self,original_candidate,indices):
        candidate = original_candidate.copy()
        del candidate[[i for i in range(len(candidate)) if i not in indices]]
        distances_abs = candidate.get_all_distances(mic=True)

        numbers = candidate.get_atomic_numbers()
        r = [covalent_radii[number] for number in numbers]
        x,y = np.meshgrid(r,r)
        optimal_distances = x+y
    
        distances_rel = distances_abs / optimal_distances
    
        return distances_rel

    def get_bond_matrix(self,candidate):
        dist = self.get_distances(candidate,self.indices)
        matrix = np.logical_and(dist > 1e-3, dist < self.covalent_bond_scale_factor).astype(int)
        matrix += np.diag(candidate.get_atomic_numbers()[self.indices])
        return matrix

    def convert_matrix_to_eigen_value_string(self,matrix):
        w,_ = np.linalg.eig(matrix)
        w = np.real(w)
        w.sort()
        s = '[' + ','.join(['{:8.3f}'.format(e) for e in w]) + ']'
        s = s.replace('-0.000',' 0.000')
        return s

    def get_eigen_value_string(self, t, allow_read_from_meta_information=False):
        # this is conditioned that the meta information is trusted (corrupted when Candidates are copied and changed)
        if allow_read_from_meta_information:
            if t.has_meta_information('eigen_string'):
                return t.get_meta_information('eigen_string')

        if self.indices is None:
            self.indices = list(range(len(self.template),len(t)))
            print('SETTING INDICES:',self.indices)

        m = self.get_bond_matrix(t)
        if self.debug:
            print(m)
        s = self.convert_matrix_to_eigen_value_string(m)
        #if t.has_meta_information('eigen_string'):
        #    assert s == t.get_meta_information('eigen_string'),'eigen string has changed: A:{}, B:{}, C:{}'.format(s,t.get_meta_information('eigen_string'),m)
        t.add_meta_information('eigen_string',s)
        return s


import os
import pytest 
from agox.test.test_utils import TemporaryFolder

@pytest.mark.ray
def test_local(tmp_path):
    import agox.test.model_tests.descriptors_api
    import agox.test.model_tests.sgpr_api
    import agox.test.model_tests.gpr_api
    import agox.test.model_tests.load_api
    import agox.test.model_tests.sgpr_api_forces

# @pytest.mark.ray
# def test_local_forces(tmp_path):
#     with TemporaryFolder(tmp_path):
#         import agox.test.model_tests.sgpr_api_forces

# @pytest.mark.ray
# def test_global(tmp_path):
#     with TemporaryFolder(tmp_path):
#         import agox.test.model_tests.gpr_api

# @pytest.mark.ray
# def test_descriptors(tmp_path):
#     with TemporaryFolder(tmp_path):
#         import agox.test.model_tests.descriptors_api



import numpy as np
from ase.data import covalent_radii
from agox.postprocessors.ABC_postprocess import PostprocessBaseClass

class MinimumDistPostProcess(PostprocessBaseClass):

    """
    Checks if all atoms in a candidate are within a certain distance of each other.

    Parameters
    ----------
    c1 : float, optional
        Minimum distance as a fraction of the sum of the covalent radii, by default 0.75.
    c2 : float, optional
        Maximum distance as a fraction of the sum of the covalent radii, by default 1.25.
    fragmented_check : bool, optional
        Check if the candidate is fragmented, by default True.
    backbone_check : bool, optional
        Check if the backbone of the candidate is fragmented, by default False.
    """

    name = 'MinimumDistance'
    def __init__(self, c1=0.75, c2=1.25, fragmented_check=True, backbone_check=False, **kwargs):
        super().__init__(**kwargs)
        self.c1 = c1
        self.c2 = c2
        self.backbone_check = backbone_check
        self.fragmented_check = fragmented_check
    
    def postprocess(self, candidate):
        if candidate is None:
            return None

        distances = self.get_distances(candidate)
        if np.any(distances[np.tril_indices(distances.shape[0], k=-1)] < self.c1):
            self.writer('Candidate minimum distance violated')
            return None

        # check max distance through laplacian spectrum
        if self.fragmented_check:
            L = self.get_laplacian_matrix(distances)
            w,_ = np.linalg.eig(L)
            if np.sum(np.abs(w) < 1e-12) > 1:
                self.writer('Candidate is fragmented')
                return None


        if self.backbone_check:
            candidate_no_H = candidate.copy()
            indices = np.argwhere(candidate_no_H.numbers == 1)
            del candidate_no_H[indices]

            distances = self.get_distances(candidate_no_H)
            L = self.get_laplacian_matrix(distances)
            w,_ = np.linalg.eig(L)

            if np.sum(np.abs(w) < 1e-12) > 1:
                self.writer('Candidate backbone is fragmented')
                return None
                                       
        self.writer('Candidate dist OK')
        return candidate


    def get_laplacian_matrix(self, relative_dist):
        A = np.logical_and(relative_dist > self.c1, relative_dist < self.c2).astype(int)
        D = np.diag(np.sum(A, axis=1))
        return D-A

    def get_distances(self, candidate):
        dist = candidate.get_all_distances(mic=True)
        cov_dist = np.array([covalent_radii[n] for n in candidate.numbers])
        relative_dist = dist / np.add.outer(cov_dist, cov_dist)
        return relative_dist

    
if __name__ == '__main__':
    from ase.build import molecule
    from ase import Atoms
    from ase.visualize import view
    h20 = molecule('H2O')
    h20 += Atoms('H', positions=[[0.,.7,.7]])
    # h20 += Atoms('C', positions=[[0.,1.4,1.4]])
    view(h20)
    print(h20.positions)
    min_dist = MinimumDistPostProcess(backbone_check=True)
    min_dist.postprocess(h20)
        



import numpy as np
from copy import copy
from agox.models.descriptors.ABC_descriptor import DescriptorBaseClass


class SOAP(DescriptorBaseClass):

    name = "SOAP"
    descriptor_type = "local"

    """
    SOAP descriptor wrapper for DScribe.

    Parameters
    ----------
    r_cut : float
        Cutoff radius for the local environment.
    nmax : int
        Maximum number of radial basis functions.
    lmax : int
        Maximum degree of spherical harmonics.
    sigma : float
        Standard deviation of the gaussian used to expand the atomic density.
    weight : bool or dict
        If True, use polynomial weighting function. If False, use no weighting
        function. If dict, use custom weighting function.
    periodic : bool
        Whether to use periodic boundary conditions.
    dtype : str
        Data type of the output.
    crossover : bool
        Whether to use crossover compression.
    compression : dict
        Compression parameters. If None, crossover is used.
    """

    def __init__(
        self,
        r_cut=4,
        nmax=3,
        lmax=2,
        sigma=1.0,
        weight=True,
        periodic=True,
        dtype="float64",
        crossover=False,
        compression=None,
        *args,
        **kwargs
    ):
        super().__init__(*args, **kwargs)

        from dscribe.descriptors import SOAP as dscribeSOAP
        self.r_cut = r_cut

        # Weighting function
        if weight is True:
            weighting = {"function": "poly", "r0": r_cut, "m": 2, "c": 1}
        elif weight is None:
            weighting = None
            method = 'analytical'
        elif weight is False:
            weighting = None
            method = 'analytical'
        else:
            weighting = weight

        # Compression / Crossover: 
        if compression is None:
            assert type(crossover) == bool, "crossover must be bool"
            if crossover:
                compression = {"mode":"crossover"}
            else:
                compression = {"mode":"off"}
        elif type(compression) == dict:
            pass        
        
        # Create dscribe SOAP descriptor object: 
        self.soap = dscribeSOAP(
            species=self.environment.get_species(),
            periodic=periodic,
            r_cut=r_cut,
            n_max=nmax,
            l_max=lmax,
            sigma=sigma,
            weighting=weighting,
            dtype=dtype,
            compression=compression,
            sparse=False,
        )

        self.length = self.soap.get_number_of_features()

    def create_features(self, atoms) -> np.ndarray:
        """
        Calculate SOAP features. 

        Parameters
        ----------
        atoms : ase.Atoms
            Atoms object.
        
        Returns
        -------
        features : np.ndarray
            Array of shape [n_centers, n_features]
        """
        return self.soap.create(atoms)

    def create_feature_gradient(self, atoms) -> np.ndarray:
        """
        Calculate the gradient of the SOAP features wrt. atomic coordinates. 

        Parameters
        ----------
        atoms : ase.Atoms
            Atoms object.
        
        Returns
        -------
        feature gradient: np.ndarray
            Array of shape [n_centers, n_atoms, n_features, 3]    
        """
        f_deriv = self.soap.derivatives(
            atoms, return_descriptor=False, attach=True, method="numerical"
        )
        return f_deriv

    def get_number_of_centers(self, atoms):
        """
        Get the number of centers for the SOAP descriptor.

        Parameters
        ----------
        atoms : ase.Atoms
            Atoms object.
        
        Returns
        -------
        n_centers : int
            Number of centers.    
        """
        return len(atoms)

    @classmethod
    def from_species(cls, species, **kwargs):
        """
        Create a SOAP descriptor from a list of species.

        Parameters
        ----------
        species : list
            List of species.
        **kwargs : dict
            Keyword arguments for the SOAP descriptor.
        """
        from ase import Atoms
        from agox.environments import Environment

        environment = Environment(
            template=Atoms(""),
            symbols="".join(species),
            use_box_constraint=False,
            print_report=False,
        )
        return cls(environment=environment, **kwargs)



from agox.samplers.ABC_sampler import SamplerBaseClass
import numpy as np
from agox.writer import agox_writer
from agox.observer import Observer

class MetropolisSampler(SamplerBaseClass):

    """
    Metropolis Sampler for selecting candidates based on the Metropolis algorithm.

    The probability of accepting a candidate is given by

    .. math::
        P = e^{-(E_{new} - E_{old})/kT}

    Typically used in the context of a Basin-Hopping algorithm.

    Parameters
    ----------
    temperature : float
        The selection temperature in eV, so it is really $kT$.
    """

    name = 'MetropolisSampler'

    def __init__(self, temperature=1, gets={'get_key':'evaluated_candidates'},
        **kwargs):
        """
        Temperature: The selection temperature in eV. 
        """
        super().__init__(gets=gets, **kwargs)

        self.temperature = temperature

        self.reset_observer() # We remove observers added by the base-class. 
        self.add_observer_method(self.setup_sampler, gets=self.gets[0], sets=self.sets[0], order=self.order[0], handler_identifier='AGOX')

    def get_random_member(self):
        if len(self.sample) > 0:
            return [self.sample[0].copy()]
        else:
            return []

    @agox_writer
    @Observer.observer_method
    def setup_sampler(self, state):
        if self.do_check():            
            evaled_candidates = state.get_from_cache(self, self.get_key)
            evaled_candidates = list(filter(None, evaled_candidates))
            if len(evaled_candidates) > 0:
                best_candidate = evaled_candidates[np.argmin([atoms.get_potential_energy() for atoms in evaled_candidates])]
                self.setup(best_candidate)

    def setup(self, potential_step):
        if isinstance(potential_step, list):
            potential_step = potential_step[-1]

        # If the current sample is not empty
        if len(self.sample) > 0:
            chosen_candidate = self.sample[0]
        else: # If it is.
            chosen_candidate = None

        # If the there is no new candidate to consider we return. 
        if potential_step is None:
            return

        # If the current sample was empty we just take the new possiblility.
        if chosen_candidate is None:
            chosen_candidate = potential_step
            if chosen_candidate is not None:
                chosen_candidate.add_meta_information('accepted', True)
        else: # Otherwise we look at the probabilities.
            Eold = chosen_candidate.get_potential_energy()
            Enew = potential_step.get_potential_energy()
            if Enew < Eold: # Always accept if the step is lower in energy. 
                chosen_candidate = potential_step
                accepted = True
            else: # Calculate the probability and roll to see if we accept. 
                P = np.exp(-(Enew - Eold)/self.temperature)
                r = np.random.rand()
                if r < P:
                    accepted = True
                    chosen_candidate = potential_step
                else:
                    accepted = False
            # Updat the acceptance. 
            potential_step.add_meta_information('accepted', accepted)

        # Add to the sample.
        self.sample = [chosen_candidate]
        
        self.writer(f'Length of sample: {len(self.sample)}')


import atexit
import os
import shutil
from time import sleep
from typing import Optional

import numpy as np
import ray


def get_cpu_count(cpu_count):
    if cpu_count is None:
        try:
            cpu_count = int(os.environ["SLURM_NTASKS"])
        except:
            cpu_count = len(os.sched_getaffinity(0))
    return cpu_count


def get_memory(memory, cpu_count):
    if memory is None:
        try:
            memory = int(os.environ["SLURM_MEM_PER_NODE"] * 1e6)
        except:
            memory = cpu_count * int(2 * 1e9)
    return memory


def get_tmp_dir(tmp_dir: Optional[str]) -> str:
    """Get a suitable directory to use as Ray logging directory.

    Parameters
    ----------
    tmp_dir : Optional[str]
        Temporary directory to attempt to use. If None, a subdirectory of the
        current working directory is used.

    Returns
    -------
    str
        Ray logging directory path.
    """
    if tmp_dir is None:
        tmp_dir = os.path.join(os.getcwd(), 'ray')

    if not check_tmp_dir_length(tmp_dir):
        tmp_dir = os.path.expanduser('~') + '/tmp/ray'
        print(f'Using {tmp_dir} as Ray temporary folder because the given or '
              'generated path is too long.')

    return tmp_dir


def check_tmp_dir_length(tmp_dir: str) -> bool:
    """Check whether the directory path is short enough to be used as Ray
    logging directory.

    Ray opens a Unix socket in this directory, and the maximum path length for
    a Unix socket on Linux is 107 bytes. The path to this Unix socket is
    determined as follows, with below each field the (maximum) length of that
    field:
    `[tmp_dir]/session_%Y-%m-%d_%H-%M-%S_%F_[pid]/sockets/plasma_store`
                        4  2  2  2  2  2  6   7*
    * The pid is limited by PID_MAX_LIMIT = 4194304.

    Hence, the additional part after `tmp_dir` is 64 characters long:
    `/session_xxxx-xx-xx_xx-xx-xx_xxxxxx_xxxxxxx/sockets/plasma_store`
     1234567890123456789012345678901234567890123456789012345678901234

    Parameters
    ----------
    tmp_dir : str
        Directory path to check.

    Returns
    -------
    bool
        Whether the directory path is short enough.
    """
    return len(tmp_dir.encode('utf-8')) + 64 <= 107


def ray_startup(
    cpu_count: int = None,
    memory: int = None,
    tmp_dir: str = None,
    include_dashboard: bool = False,
    max_grace_period: float = 0.0,
    keep_log_files: bool = False,
):
    """
    Start a Ray instance on Slurm.

    Parameters
    ----------
    cpu_count : int, optional
        Number of CPU cores to use, by default None in which case defaults are
        read from SLURM.
    memory : int, optional
        Amount of memeory to use in bits, by default None in which case environment
        variables are used to guess a suitable amount.
    tmp_dir : str, optional
        Temporary directory for Ray to use, by default None in which case a
        directory is created in the current working directory.
    include_dashboard : bool, optional
        Whether or not to include the Ray dashboard, by default False.
    keep_log_files : bool, optional
        Whether to keep the Ray log files after the task is terminated, by
        default False.
    """

    if ray.is_initialized():
        return ray.cluster_resources()

    # CPU Count:
    cpu_count = get_cpu_count(cpu_count)
    memory = get_memory(memory, cpu_count)
    tmp_dir = get_tmp_dir(tmp_dir)

    if max_grace_period > 0:
        grace = np.random.uniform(0, max_grace_period)
        print(f'Sleeping {grace} s', flush=True)
        sleep(grace)
    ray_context = ray.init(
        address='local',
        object_store_memory=int(memory / 4),
        num_cpus=cpu_count,
        ignore_reinit_error=True,
        include_dashboard=include_dashboard,
        _temp_dir=tmp_dir,
        _memory=memory,
    )

    ray_stats = ray.cluster_resources()
    print("Ray Resources")
    for key, value in ray_stats.items():
        print(f"\t{key} = {value}")

    if not keep_log_files:
        atexit.register(ray_cleanup, path=ray_context.address_info.get('session_dir'))

    return ray_stats


def ray_cleanup(path: str):
    """Shut down Ray and (try to) delete the log directory.

    Parameters
    ----------
    path : str
        Path to the log directory.
    """
    ray.shutdown()

    if os.path.exists(path):
        try:
            shutil.rmtree(path)
        except Exception as e:
            print('Exception occurred when cleaning up the Ray log directory:')
            print(str(e))



import os
from agox.module import Module
from pathlib import Path
import logging
import time

class TestModule(Module):
    name = "TestModule"

    def __init__(self):
        super().__init__()

    def run(self, extra_str):
        print(f"Hello world: {extra_str}!")

def remote_module_function(test_module, *args, **kwargs):
    test_module.run(*args)
    return 0

def remote_function(*args, **kwargs):
    print(f"Hello world: {args}!")
    return 0

def log_slurm_info(logger):
    logging.info("SLURM environment variables:")
    for var, value in os.environ.items():
        if 'SLURM' in var:
            logger.info(f"{var}: {value}")

    logging.info("")

def test_stability(logger):
    os.environ["RAY_DEDUP_LOGS"] = "0"
    from agox.utils.ray import ray_startup, get_ray_pool

    logger.info('Event 1: Stability test started')

    # Start Ray
    ray_startup()
    logger.info('Event 2: Ray started')

    # Make a module:
    module = TestModule()
    logger.info('Event 3: Module created')

    # Get the ray pool
    pool, _ = get_ray_pool()
    logger.info('Event 4: Ray pool created')

    print(f"Pool: {pool}")
    pool.execute_on_actors(remote_function, [], ['1'], {})
    logger.info('Event 5: Function executed')
    time.sleep(2)

    # Add the module to the pool
    pool.add_module(module)
    logger.info('Event 6: Module added to pool')

    # Run the module
    pool.execute_on_actors(remote_module_function, [module.ray_key], ['1'], {})
    logger.info('Event 7: Module executed')

def slurm_test():
    
    submit_dir = os.getenv('SLURM_SUBMIT_DIR')
    array_task_id = os.getenv('SLURM_ARRAY_TASK_ID')

    if submit_dir is not None and array_task_id is not None:
        path = Path(submit_dir) / f"{array_task_id}/ray.log"
    else:
        path = Path('ray.log')

    print(f"Logging to: {path.resolve()}")

    logger = logging.getLogger(__name__)
    logging.basicConfig(filename=path, level=logging.DEBUG, format='%(asctime)s - %(message)s')
    logging.info('Event 0: Log file created.')

    log_slurm_info(logger)
    test_stability(logger)




import numpy as np
from agox.utils.filters.ABC_filter import FilterBaseClass
from typing import List, Tuple
from ase import Atoms
from agox.models.descriptors.ABC_descriptor import DescriptorBaseClass


class Cluster:
    def __init__(self, representative: str):
        self.representative = representative
        self.structures = []
        self.energies = []
        self.features = []
        self.indices = []

    def add_member(self, atoms: Atoms, index: int, feature: np.ndarray):
        self.indices.append(index)
        self.structures.append(atoms)
        self.features.append(feature)

        try:
            energy = atoms.get_potential_energy()
        except:
            energy = 0
        self.energies.append(energy)

        if self.representative == "energy":
            self.rep_index = np.argmin(self.energies)
        elif self.reepresentative == "first":
            self.rep_index = 0

    def get_representative_member(self) -> Atoms:
        return self.structures[self.rep_index]

    def get_representative_feature(self) -> np.ndarray:
        return self.features[self.rep_index]

    def get_representative_index(self) -> int:
        return self.indices[self.rep_index]


class FeatureDistanceFilter(FilterBaseClass):
    name = "FeatureDistanceFilter"

    """
    Filter that removes duplicates structures based on the distance between their features.

    Parameters
    ----------
    descriptor: DescriptorBaseClass
        Descriptor used to calculate the features.
    threshold: float
        Maximum distance between features.
    representative: str
        Method used to select the representative structure.
        Options: "energy", "first"
    """

    def __init__(
        self,
        descriptor: DescriptorBaseClass,
        threshold: float,
        representative: str = "energy",
    ):
        super().__init__()
        self.descriptor = descriptor
        self.threshold = threshold

        assert representative in ["energy", "first"]
        self.representative = representative

    def _filter(self, atoms: List[Atoms]) -> np.ndarray:
        # Sum over atoms, converts local descriptors to global descriptors
        # Doesnt change global descriptors.
        F = np.array(self.descriptor.get_features(atoms)).sum(axis=1)

        clusters = []
        for i, atoms in enumerate(atoms):
            added_to_cluster = False
            for cluster in clusters:
                cluster_feature = cluster.get_representative_feature()
                if np.linalg.norm(F[i] - cluster_feature) < self.threshold:
                    cluster.add_member(atoms, i, F[i])
                    added_to_cluster = True
                    break

            if not added_to_cluster:
                new_cluster = Cluster(representative=self.representative)
                new_cluster.add_member(atoms, i, F[i])
                clusters.append(new_cluster)

        indices = [c.get_representative_index() for c in clusters]
        return indices



import matplotlib
matplotlib.use("Agg")

import numpy as np
from ase import Atoms
from ase.io import read
from ase.optimize import BFGS

from agox import AGOX
from agox.acquisitors import MetaInformationAcquisitor
from agox.collectors import ParallelTemperingCollector
from agox.databases import Database
from agox.environments import Environment
from agox.evaluators import LocalOptimizationEvaluator
from agox.samplers import ParallelTemperingSampler
from agox.postprocessors import ParallelRelaxPostprocess

from agox.models.descriptors.fingerprint import Fingerprint
from agox.models.GPR.kernels import RBF, Constant as C, Noise
from agox.models.GPR import GPR
from agox.models.GPR.priors import Repulsive

seed = 42
database_index = 0

sample_size = 10
rattle_amplitudes = np.linspace(0.1, 5, sample_size)

################################################################################
# Calculator
##############################################################################

from ase.calculators.emt import EMT
calc = EMT()

##############################################################################
# System & general settings:
##############################################################################

db_path = "db{}.db".format(database_index)
database = Database(filename=db_path, order=6)

template = Atoms("", cell=np.eye(3) * 12)
confinement_cell = np.eye(3) * 6
confinement_corner = np.array([3, 3, 3])
environment = Environment(
    template=template,
    symbols="Au12Ni2",
    confinement_cell=confinement_cell,
    confinement_corner=confinement_corner,
)

##############################################################################
# Search Settings:
##############################################################################

# Setup a ML model.
descriptor = Fingerprint(environment=environment)
beta = 0.01
k0 = C(beta, (beta, beta)) * RBF()
k1 = C(1 - beta, (1 - beta, 1 - beta)) * RBF()
kernel = C(5000, (1, 1e5)) * (k0 + k1) + Noise(0.01, (0.01, 0.01))
model = GPR(descriptor=descriptor, kernel=kernel, database=database, prior=Repulsive())

sampler = ParallelTemperingSampler(
    t_min=0.02, t_max=2, swap="down", sample_size=sample_size, order=3, verbose=True,
    model=model)

random_generator_kwargs = dict(contiguous=True)

collector = ParallelTemperingCollector.from_sampler(
    sampler, environment, rattle_amplitudes, random_generator_kwargs=random_generator_kwargs, dynamic=False)

relaxer = ParallelRelaxPostprocess(
    model=model,
    constraints=environment.get_constraints(),
    optimizer_run_kwargs={"steps": 5, "fmax": 0.1}, 
    start_relax=0,
    optimizer=BFGS,
    order=2,
)

# Each iteration the lowest temperature walker/replica will be evaluated.
# Skip chance can be used to skip evaluations with some probability.
acquisitor = MetaInformationAcquisitor(
    meta_key="walker_index", skip_chance=0.0, order=4
)

evaluator = LocalOptimizationEvaluator(
    calc,
    number_to_evaluate=1,
    optimizer_kwargs={"logfile": None},
    optimizer_run_kwargs={"fmax": 0.05, "steps": 0},
    constraints=environment.get_constraints(),
    order=5,
)

agox = AGOX(
    collector,
    relaxer,
    sampler,
    acquisitor,
    evaluator,
    database,
    seed=seed,
)

# ##############################################################################
# # Lets get the show running!
# ##############################################################################

agox.run(N_iterations=10)



import numpy as np
from abc import ABC, abstractmethod

class DataSelector(ABC):

    def __init__(self, replace=False):
        self.replace = replace

    def __call__(self, data, N, return_indices=False):
        p = self.make_probability(data, N)
            
        indices = np.random.choice(np.arange(len(data)), p=p, size=N, replace=self.replace)
        selected_data = [data[i] for i in indices]
        if return_indices:
            return selected_data, indices
        return selected_data

    @abstractmethod
    def make_probability(self, data, N):
        pass

class UniformSelector(DataSelector):

    def make_probability(self, data, N):
        return np.ones(len(data)) * 1 / len(data)

class SegmentedSelector(DataSelector):

    def make_probability(self, data, N):
        p = np.zeros(len(data))
        energies = np.array([atoms.get_potential_energy() for atoms in data])
        sorted_indices = np.argsort(energies)
        interval = int(np.floor(len(data) / N))
        p[sorted_indices[0:-1:interval]] = 1
        return p / np.sum(p)

class EnergyUniformSelector(DataSelector):

    def __init__(self, max_delta, bin_width=1, **kwargs):
        super().__init__(**kwargs)
        self.max_delta = max_delta
        self.bin_width = bin_width

    def make_probability(self, data, N):
        energies = np.array([a.get_potential_energy() for a in data])

        e_min = energies.min()
        e_max = np.min([e_min+self.max_delta, energies.max()])
        bins = np.arange(e_min, e_max, self.bin_width) + self.bin_width / 2
        bin_count, bin_indices = self.get_bin_count(bins, energies)

        effective_nbins = np.count_nonzero(bin_count!=0)
        bin_prob = 1 / (effective_nbins - 1)

        bin_probs = np.zeros_like(bins)
        for i, bc in enumerate(bin_count):
            if bc != 0:
                p = bin_prob / bc
            else:
                p = 0
            bin_probs[i] = p

        bin_probs[-1] = 0

        probabilities = np.zeros_like(energies)
        for i, bin_index in enumerate(bin_indices):
            probabilities[i] = bin_probs[bin_index]

        return probabilities
        
    def get_bin_count(self, bins, energies):
        bin_count = np.zeros_like(bins).astype(int)
        bins = bins.copy().reshape(-1, 1)
        energies = energies.copy().reshape(1, -1)
        indices = np.argmin(np.abs(bins - energies), axis=0)
        for index in indices: 
            bin_count[index] += 1

        return bin_count, indices

def split(data, fractions):
    N_total = len(data)
    counts = [int(frac * N_total) for frac in fractions]
    counts[-1] += N_total - np.sum(counts)
    counts = [0] + counts
    counts = np.cumsum(counts)

    indices = np.arange(N_total)
    np.random.shuffle(indices)

    splits = []
    for i in range(len(fractions)):
        splits.append([data[j] for j in indices[counts[i]:counts[i+1]]])

    return splits

def exclude(data, indices):
    return [data[i] for i in range(len(data)) if i not in indices]

def filter_energy(data, max_delta):
    energies = np.array([a.get_potential_energy() for a in data])
    indices = np.argwhere(energies < energies.min()+max_delta).flatten()
    return [data[i] for i in indices]



        






from typing import Dict, Tuple, Union

import numpy as np
from ase.atoms import Atoms
from ase.geometry import get_distances
from numpy.typing import NDArray

from agox.candidates.standard import StandardCandidate
from agox.models.descriptors.voronoi import Voronoi

SiteMapping = Dict[Tuple[float, float], str]

fcc111_mapping = {
    (0.0, 0.0): 'ontop',
    (0.5, 0.0): 'bridge',
    (0.0, 0.5): 'bridge',
    (0.5, 0.5): 'bridge',
    (1/3, 1/3): 'fcc',
    (2/3, 2/3): 'hcp'
}

site_mappings = {
    'fcc111': fcc111_mapping
}


class VoronoiSite(Voronoi):

    """Adsorption site-aware Voronoi graph descriptor.

    Parameters
    ----------
    site_mapping : Union[SiteMapping, str]
        Mapping from scaled unit cell coordinates (x, y) to special adsorption
        sites, or string representing predefined mappings from `site_mappings`.
    """

    name = 'VoronoiSite'

    def __init__(self,
                 *args,
                 site_mapping: Union[SiteMapping, str],
                 **kwargs):
        super().__init__(*args, **kwargs)

        if isinstance(site_mapping, str):
            if site_mapping not in site_mappings:
                raise KeyError(f'{site_mapping} is not a valid site mapping key. '
                               f'Valid keys: {", ".join(site_mappings.keys())}')
            site_mapping: SiteMapping = site_mappings.get(site_mapping)

        self.site_mapping = site_mapping

    def get_bond_matrix(self, candidate: StandardCandidate) -> NDArray:
        """Calculate the bond matrix for a candidate structure.

        Parameters
        ----------
        candidate : StandardCandidate
            Candidate structure.

        Returns
        -------
        NDArray
            Bond matrix.
        """
        voronoi_bond_matrix = super().get_bond_matrix(candidate)

        adsorbate = Atoms(candidate)[candidate.get_optimize_indices()]
        labels = self._assign_labels(adsorbate)

        diagonal = np.zeros(len(self.indices))
        for label_idx, label in zip(candidate.get_optimize_indices(), labels):
            if label_idx in self.indices:
                diagonal[self.indices.index(label_idx)] = label

        bond_matrix = voronoi_bond_matrix + np.diag(diagonal)
        return bond_matrix

    def _assign_labels(self, adsorbate: Atoms) -> NDArray:
        """Assign adsorption site labels to the adsorbate atoms.

        The adsorption site labels :math:`s_i \in [0, 1)` encode which standard
        adsorption site each adsorbate atom is closest to.

        Parameters
        ----------
        adsorbate : Atoms
            Adsorbate atoms object.

        Returns
        -------
        NDArray
            Adsorption site labels.
        """
        adsorbate_info: Dict = self.template.info.get('adsorbate_info')
        if adsorbate_info is None:
            raise ValueError('Template was not originally created with `ase.build`')

        cell: NDArray = adsorbate_info.get('cell')

        site_positions = np.asarray(list(self.site_mapping.keys()))
        site_types = list(self.site_mapping.values())
        unique_site_types = sorted(set(site_types))

        # convert cell into 3D
        cell = np.array([[cell[0, 0], cell[0, 1], 0.0],
                         [cell[1, 0], cell[1, 1], 0.0],
                         [0.0, 0.0, 1.0]])

        adsorbate = adsorbate.copy()
        adsorbate_positions = adsorbate.get_positions()
        adsorbate_positions[:, 2] = 0.0  # we don't care about the z coordinates

        site_positions = np.hstack((site_positions, np.zeros((site_positions.shape[0], 1))))  # convert into 3D
        site_positions = site_positions @ cell

        _, distances = get_distances(adsorbate_positions, site_positions, cell=cell, pbc=[True, True, False])

        site_indices = np.argmin(distances, axis=1)
        adsorbate_labels = (np.asarray([unique_site_types.index(site_types[i]) for i in site_indices], dtype=int) + 1) / \
            (len(unique_site_types) + 1)

        return adsorbate_labels



import numpy as np
from agox.acquisitors.ABC_acquisitor import AcquisitorBaseClass, AcquisitonCalculatorBaseClass
from ase.calculators.calculator import all_changes

class LowerConfidenceBoundAcquisitor(AcquisitorBaseClass):

    """
    Lower Confidence Bound Acquisitor, using predictions and uncertainties from a model.
    
    The acquisition function is given by

    .. math::

        F(x) = E(x) - \kappa \sigma(x)

                
    Parameters
    -----------

    model: agox.models.ABC_model.ModelBaseClass
        Model that is used for the acquisition function.
    kappa: float
        Constant that is multiplied with the uncertainty.
    kwargs: dict
        Dictionary of keyword arguments passed to the AcquisitorBaseClass.        
    """

    name = 'LCBAcquisitor'

    def __init__(self, model, kappa=1, **kwargs):
        super().__init__(**kwargs)
        self.kappa = kappa
        self.model = model

    def calculate_acquisition_function(self, candidates):
        fitness = np.zeros(len(candidates))
        
        # Attach calculator and get model_energy
        for i, candidate in enumerate(candidates):
            E, sigma = self.model.predict_energy_and_uncertainty(candidate)
            fitness[i] = self.acquisition_function(E, sigma)

            # For printing:
            candidate.add_meta_information('model_energy', E)
            candidate.add_meta_information('uncertainty', sigma)

        return fitness

    def print_information(self, candidates, acquisition_values):
        if self.model.ready_state:
            for i, candidate in enumerate(candidates):
                fitness = acquisition_values[i]
                Emodel = candidate.get_meta_information('model_energy')
                sigma = candidate.get_meta_information('uncertainty')
                generator = candidate.get_meta_information('generator')
                self.writer('Candidate: E={:8.3f}, s={:8.3f}, F={:8.3f}, {}'.format(Emodel, sigma, fitness, generator))

    def get_acquisition_calculator(self):
        return LowerConfidenceBoundCalculator(self.model, self.acquisition_function, self.acquisition_force)

    def acquisition_function(self, E, sigma):
        return E - self.kappa * sigma

    def acquisition_force(self, E, F, sigma, sigma_force):
        return F - self.kappa*sigma_force

    def do_check(self, **kwargs):
        return self.model.ready_state

class LowerConfidenceBoundCalculator(AcquisitonCalculatorBaseClass):

    implemented_properties = ['energy', 'forces']

    def __init__(self, model, acquisition_function, acquisition_force, **kwargs):
        super().__init__(model, **kwargs)
        self.acquisition_function = acquisition_function
        self.acquisition_force = acquisition_force

    def calculate(self, atoms=None,
                  properties=['energy'],
                  system_changes=all_changes):
        super().calculate(atoms, properties, system_changes)

        derivatives = 'forces' in properties
        model_data = self.model.converter(atoms, derivatives=derivatives)
        
        E = self.model.predict_energy(atoms, **model_data)
        sigma = self.model.predict_uncertainty(atoms, **model_data)
        self.results['energy'] = self.acquisition_function(E, sigma)

        if derivatives:

            F = self.model.predict_forces(atoms, **model_data)
            sigma_force = self.model.predict_uncertainty_forces(atoms, **model_data)
        
            self.results['forces'] = self.acquisition_force(E, F, sigma, sigma_force)


        




"""
An environment describes the search problem
- How many and which types of atoms will the search place? 
- What is the computational cell? 
- Is there a fixed template, e.g. a surface to be decorated by the search? 
- Is the search constrained to a certain region of space?

These are settings that are not part of the search algorithm itself, but 
rather describe the problem that the search is trying to solve.
"""
from .ABC_environment import EnvironmentBaseClass
from agox.environments.environment import Environment

__all__ = [
    'EnvironmentBaseClass',
    'Environment']


import numpy as np
from ase.optimize import BFGS
from numpy.linalg import eigh

class SafeBFGS(BFGS):

    def step(self, f=None):
        atoms = self.atoms

        if f is None:
            f = atoms.get_forces()

        r = atoms.get_positions()
        f = f.reshape(-1)
        self.update(r.flat, f, self.r0, self.f0)
        omega, V = eigh(self.H)

        if not np.all(np.fabs(omega) > 0):
            # Hessian is singular - Probably due a bad step taken with very large 
            # force. This just resets the Hessian. 
            self.H = np.eye(3 * len(self.atoms)) * self.alpha
            omega, V = eigh(self.H)

        dr = np.dot(V, np.dot(f, V) / np.fabs(omega)).reshape((-1, 3))
        steplengths = (dr**2).sum(1)**0.5
        dr = self.determine_step(dr, steplengths)
        atoms.set_positions(r + dr)
        self.r0 = r.flat.copy()
        self.f0 = f.copy()
        self.dump((self.H, self.r0, self.f0, self.maxstep))




"""
This module contains the filters that can be used to filter candidate (or ase.Atoms) objects. 
This is useful for filtering out candidates that are not interesting, e.g. if the energy is too high.
"""


from .ABC_filter import FilterBaseClass
from .all import AllFilter
from .none import NoneFilter
from .energy import EnergyFilter
from .ABC_filter import FilterBaseClass, SumFilter
from .kmeans_energy import KMeansEnergyFilter
from .sparse_filter import SparsifierFilter
from .random import RandomFilter
from .voronoi import VoronoiFilter

__all__ = ['FilterBaseClass',
            'AllFilter',
            'NoneFilter',
            'EnergyFilter',
            'SumFilter',
            'KMeansEnergyFilter',
            'SparsifierFilter',
            'RandomFilter',
            'VoronoiFilter',
        ]